/* 
 * Capsule.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.CommQueuePuller;
import edu.utah.janos.nodeos.CommSpaceElement;
import edu.utah.janos.io.Log;

/**
 * A capsule is a combination of a packet and its forwarding routine;
 * the forwarding routine is executed at every active node the
 * packet visits in the network.
 * <p>
 * This is an abstract, an actual Capsule is created by subclassing
 * this class and filling in the missing routines.
 * <p>
 * Subclasses must implement the mid(), pid(), and evalutate() methods.
 * <p>
 * All ANTS capsules have the following format:
 * <pre>
 *   <ANEP> header
 *   4-byte ANTS version number
 *   16-byte MethodID  (The MD5 hash over the class bytecode)
 *   4-byte logical source address
 *   4-byte logical destination address
 *   4-byte logical previous-hop address
 *   4-byte TTL field.
 *   ... Capsule-specific data ...
 * </pre>
 *
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public abstract class Capsule {
    static final byte ANEPVERSION = 1;
    static final short ANTSTYPE = 18; // MIT ANTS type
    static final short ANTSRTYPE = 121; // Utah ANTSR type
    static final short SRCOPTTYPE = 1;
    static final short DESTOPTTYPE = 2;
    static final int IPV4SCHEME = 1;
    static final int ANTS_VERSION = 0x00020004;
    static final int HEADERLENGTH =
	Xdr.TYPEID + Xdr.INT + Xdr.INT + Xdr.INT + Xdr.INT;
    
    public static final int MAX_TTL = 255;
    
    /**
     * Return the MethodID associated with this capsule.
     * <p>
     * See <a href="On_IDs.txt">On_IDs.txt</a> for more info on
     * MethodIDs.
     *
     * @see #findMID(java.lang.String)
     * @see Node#findMID(java.lang.String)
     * @return the byte[] representing the MD5 checksum.
     */
    protected abstract byte[] mid();
    
    /**
     * Return the ProtocolID associated with this capsule.
     * <p>
     * See <a href="On_IDs.txt">On_IDs.txt</a> for more info on
     * ProtocolIDs.
     *
     * @see #findPID(java.lang.String)
     * @return the byte[] representing the MD5 checksum.
     */
    protected abstract byte[] pid();
    
    /**
     * All capsules originate from some source node. 
     * This is the 32-bit IP address of that source.
     * <p>
     * The source is set implicitly by the node that
     * injects the capsule into the network and is
     * never changed after that.
     * <p>
     * XXX what is this used for by the Node?  Or is
     * it just a convenience for subclasses?  If that's
     * all then nuke it.
     *
     * @see #getSrc()
     */
    private int src;
    
    /**
     * The destination for this capsule.  Forwarding code
     * may read/write this field.
     *
     * @see #getDst()
     * @see #setDst(int)
     */
    private int dst;
    
    /**
     * 32-bit IP address of previous active node.  This
     * is used by the code loading protocol to dynamically find
     * the code to process a new packet type.
     * <p>
     * XXX security implications of giving this package access?
     *
     * @see #getPrevious()
     * @see #setPrevious(int)
     */
    private int previous; // accessed directly by channel
    
    
    /**
     * The xdr stream representing the buffer that
     * underlies a Capsule.
     * <p>
     * For incoming packets,
     * this field is set by the Protocol receive()
     * function.
     * <p>
     * For outgoing packets, this field is set
     * by the encode() method which creates a new
     * Xdr.
     */
    private Xdr xdr;
    
    /**
     * Time to live for a packet.  Used to control
     * packet propagation.
     * <p>
     * XXX a byte would suffice?
     *
     * @see #getTTL()
     */
    private int ttl;
    
    /**
     * Create a dull capsule.  
     */
    public Capsule() 
    { 
    }
    
    /**
     * Get the 32-bit IP address of the active node
     * this packet just came from.  Used to find
     * code to process this packet.
     */
    public int getPrevious()
    {
	return previous;
    }
    
    /**
     * Set the 32-bit IP address of the active node
     * this packet "came from".  Used by the outgoing
     * encode() operation.
     *
     * @see #encodeCapsuleForSend(Node)
     */
    void setPrevious(int previous) 
    {
	this.previous = previous;
    }
    
    
    /**
     * Return the TTL-like counter for this capsule.
     *
     * @return a TTL value for this Capsule.
     */
    public int getTTL()
    {
	return ttl;
    }
    
    /**
     * Access point to initialze capsule methodIDs.  Given
     * Capsule class name, returns the appropriate MID.
     * <p>
     * XXX its probably bad that the actual underlying
     * byte[] is returned and not a copy...  Who uses this?
     *
     * @param name the name of the Capsule class
     * @return the MD5 checksum (MID) associated with the given Capsule class.
     * @see Node#findMID(java.lang.String)
     */
    public static byte[] findMID(String name)
    {
	// XXX return a *copy* of the byte array???
	
	Assert.assert(Node.currentNode());
	
	TypeID t = Node.currentNode().findMID(name);
	if (t == null)
	    return null;
	return t.unwrap();
    }
    
    /**
     * Access point to initialze capsule protocolID.  Given
     * Capsule class name, returns the appropriate PID.
     *
     * @param name the name of the Capsule class
     * @return the MD5 checksum (PID) associated with the given Capsule class.
     * @see Node#findPID(java.lang.String)
     */
    public static byte[] findPID(String name)
    {
	// XXX return a *copy* of the byte array???
	
	Assert.assert(Node.currentNode());
	
	TypeID t = Node.currentNode().findPID(name);
	if (t == null)
	    return null;
	return t.unwrap();
    }
    
    /**
     * Find a given extension.  Given the name for an extension,
     * return the local "index" of that extension.
     * <p>
     * XXX This interface will need to be replaced to allow
     * extensions to be reference counted (and unloaded).
     *
     * @param name The name of the extension to load.
     * @return The local "index" of the given extension.
     */
    public static int findExtension(String name)
    {
	return 0;
    }
    
    /**
     * The forwarding routine for the Capsule.
     * This is invoked by the router when a capsule is
     * received.
     * <p>
     * This method should not access other Java classes,
     * and should only interact with the local system
     * through the local node interface.  This routine
     * should "terminate quickly."
     * <p>
     * XXX perhaps default to forwarding the Capsule?
     *
     * @param node the local node interface
     * @return an indicationof whether the forwarding was
     * successful.
     */
    abstract public boolean evaluate(Node node);
    
    /**
     * Return the source IP address of this capsule.
     *
     * @return the source IP address of this capsule.
     */
    public int getSrc()
    {
	return src;
    }
    
    /**
     * Return the destination IP address of this capsule.
     *
     * @return the destination IP address of this capsule.
     */
    public int getDst()
    {
	return dst;
    }
    
    /**
     * Set the destination IP address of this capsule.
     *
     * @param dst the destination IP address of this capsule.
     */
    public void setDst(int dst)
    {
	this.dst = dst;
    }
    
    /**
     * Return the length of the capsule header that
     * all capsule encodings will include.
     *
     * @return the length of the capsule header
     */
    public int length()
    {
	return HEADERLENGTH;
    }
    
    /**
     * Encode this capsule's header into a new Xdr.  Return the
     * new Xdr.
     * <p>
     * Because this header must go first in the Xdr, this
     * encoder unilaterally sets the xdr buffer pointer
     * to the beginning of the buffer and writes its
     * data there.
     * <p>
     * The format of the Xdr is:
     * <pre>
     * MethodID         (16 bytes)
     * src address      ( 4 bytes)
     * dst address      ( 4 bytes)
     * previous address ( 4 bytes)
     * TTL              ( 4 bytes)
     * <pre>
     *
     * @return the new Xdr representing this capsule.
     * @see #xdr 
     */
    public Xdr encode()
    {
	xdr.PUT(this.mid(), 0, Xdr.TYPEID);
	xdr.PUT(src);
	xdr.PUT(dst);
	xdr.PUT(previous);
	xdr.PUT(ttl);
	return xdr;
    }
    
    /**
     * Encode an outgoing capsule.
     * <p>
     * This writes the ANEP and ANTS header for an outgoing capsule in its
     * associated Xdr.  A new Xdr is created if the Capsule doesn't already
     * have one.  Then the per-capsule encoding is called.
     * <p>
     * This function should be invoked immediately before pushing
     * the capsule out the door.
     * <p>
     * XXX: rename to 'encodeForSend()'
     * 
     * @param node the node interface 
     * @return Xdr returns Xdr in which the remainder of the capsule should be encoded.
     * @exception SecurityException if the source is not defined in the capsule
     * @exception ResourceLimitException if the capsule has insufficent resources
     */
    public Xdr encodeCapsuleForSend(Node node)
    {
	short packetLen;
	
	// source, etc. must be set already
	if (this.getSrc() == 0)
	{
	    // XXX throw a real exception.
	    throw new Error("SecurityError: source not set in encode for " + this);
	}
        
	// previous is updated for use by the code loading
	this.setPrevious(node.getAddress()); 
	
	// ttl decremented here just before we leave
	if (this.ttl-- <= 0)
	{
	    throw new ResourceLimitException("ttl hit 0"); // XXX
	}
	
	Xdr xdr;
	
	packetLen = (short)(32 + Xdr.INT + this.length());
	
	// This could blow up if this.length() is a lie.
	xdr = new Xdr(packetLen);
	
	// Insert Header
	xdr.PUT(ANEPVERSION);
	xdr.PUT((byte)0);  // Forward packet if TypeID unknown
	xdr.PUT(ANTSTYPE);
	
	// Header with src and dest options = 8 32-bit words
	xdr.PUT((short)8); // header len
	xdr.PUT(packetLen);
	
	xdr.PUT(DESTOPTTYPE); // destination option
	xdr.PUT((short)3);
	xdr.PUT(IPV4SCHEME);
	xdr.PUT(this.getDst());
	
	xdr.PUT(SRCOPTTYPE); // source option
	xdr.PUT((short)3);
	xdr.PUT(IPV4SCHEME);
	xdr.PUT(this.getSrc());
	
	// End of ANEP header, starting ANTS headers
	
	/**
	 * XXX we could put the ANTS_VERSION magic into the
	 * MIDIDRequest/MIDIDResponse handshake.
	 */
	xdr.PUT(ANTS_VERSION);
	
	Assert.assert(xdr.index() == ((1 + 1 + 2) + (2 + 2) + (2 + 2 + 4 + 4) + (2 + 2 + 4 + 4) + (4)));
	
	this.xdr = xdr;
	this.encode();
	
	Assert.assert(xdr.index() == xdr.length()); // XXX indicates packet changed during encode
	
	return xdr;
    }
    
    /*package*/ void decodeWith(Xdr xdr)
    {
	Assert.assert(this.xdr == null);
	this.xdr = xdr;
	this.decode();
    }
    
    /**
     * Decode a capsule header from this Capsule's
     * xdr field.
     *
     * @return the new Xdr based on the current xdr.
     */
    public Xdr decode()
    {
	// Caller skip'd over the MID field.  If there was one...
	
	// pull out rest of ANTS headers.
	// XXX redundant with ANEP options...
	src = xdr.INT();
	dst = xdr.INT();
	previous = xdr.INT();
	ttl = xdr.INT();
	
	// return Xdr so that subclasses can continue decode.
	return xdr;
    }
    
    /**
     * Peek at the 'previous' field of buf.
     * For bootstrapping capsule code loading.
     *
     * @param buf the Buffer to peek into.
     * @return the value of the 'previous' field in buf.
     */
    final static int peekPrevious(BufferHandle buf)
    {
	// Copy 1 INT after skipping a TYPEID and two INTs.
	return buf.getInt(Xdr.TYPEID + Xdr.INT + Xdr.INT);
    }
    
    /**
     * Prime this capsule for transmission. Use the given parent
     * Capsule to prime the src and ttl of this capsule.
     * <p>
     * Does not maintain zero-sum ttl.  This capsule
     * gets parent's ttl minus 1.  Parent loses one hop.
     * 
     * @param parent the capsule to take src and ttl (minus 1)
     * from.
     */
    public final void prime(Capsule parent)
    {
	int newTtl = --parent.ttl;
	
	//if (newTtl <= 0)
	//	throw new ResourceLimitException(); // XXX
	
	prime(parent.src, newTtl);
    }
    
    /**
     * Priveleged prime operation.  Fill in this capsule's
     * src and ttl fields directly.  The given ttl is
     * capped at MAX_TTL
     * <p>
     * This is package-access only, otherwise it would be
     * a security hole.
     *
     * @param src the source address for the new packet.
     * @param ttl the tll for the new packet.
     * @see #MAX_TTL
     */
    final void prime(int src, int ttl)
    {
	Assert.require(src != 0); // XXX throw an Error
	Assert.require(ttl >= 0); // XXX throw an Error

	this.src = src;
	if (ttl > MAX_TTL)
	    ttl = MAX_TTL;
	this.ttl = ttl;
    }
    
    /**
     * Package up this capsule and dump it onto the queue for another
     * application
     */
    final void enqueueCapsule(CommSpaceElement msg[], int start)
    {
	if( this.xdr == null )
	{
	    short packetLen;
	    
	    packetLen = (short)(32 + Xdr.INT + this.length());
	    this.xdr = new Xdr(packetLen);
	}
	else
	{
	    this.xdr.reset();
	}
	
	byte packetClass[] = this.mid();
	byte packet[] = this.encode().getBytes();
	
	msg[start] = new CommSpaceElement(packetClass);
	msg[start + 1] = new CommSpaceElement(packet);
    }
    
    /**
     * Produce a capsule from data pushed through a CommQueue by
     * enqueueCapsule
     */
    final static Capsule dequeueCapsule(Node node,
					CommQueuePuller queue)
    {
	byte[] midbytes = null;
	byte[] packet = null;
	Capsule retval = null;
	boolean done = false;
	
	while( !done )
	{
	    try
	    {
		/*
		 * Get the MID and actual packet from the
		 * queue, we're being a bit anal here since its
		 * possible we could get interrupted during the
		 * waitToTake and bad things would happen.
		 */
		if( midbytes == null )
		    midbytes = queue.waitToTake().
			getByteArray();
		if( packet == null )
		    packet = queue.waitToTake().
			getByteArray();
		
		/* Lookup the class for this MID */
		TypeID mid = TypeID.wrap(midbytes);
		CapsuleClass cc = node.lookupCapsuleClass(mid);
		
		if( cc == null )
		{
		    node.log(Log.ERR,
			     "Couldn't resolve MID ("
			     + mid
			     + ") to a capsule class");
		    done = true;
		}
	
		/*
		 * Convert the packet bytes into an XDR
		 * for decoding
		 */
		Xdr xdr = new Xdr(packet);
		
		try
		{
		    Class cl = cc.cl;

		    done = true;

		    if (cl == null)
			throw new NullPointerException("Class for " +cc+ " is NULL.  That's bogus!");
		    
		    /* Skip header junk */
		    xdr.skip(16);
		    retval = Capsule.create(cl, xdr);
		}
		catch(Throwable th)
		{
		    node.log(Log.ERR,
			     "There was an error in "
			     + "decoding a packet");
		    node.logStackTrace(Log.ERR, th);
		}
	    }
	    catch(InterruptedException e)
	    {
		/* Oops... */
	    }
	}
	return retval;
    }
    
    /**
     * String representation of a Capsule.  Includes the
     * class name, source and destination.
     */
    public String toString()
    {
	return this.getClass().getName() +
	    "[" +
	    "s=" +NodeAddress.toString(src)+ ";" +
	    "d=" +NodeAddress.toString(dst)+ ";" +
	    "p=" +NodeAddress.toString(previous)+ 
	    "]";
    }
    
    /**
     * Create a new Capsule from the given Xdr.  The
     * xdr will be decoded into an instance of the
     * given class which must be a subclass of Capsule.
     *
     * @param cl the subtype of capsule to create
     * @param xdr the encoding of the actual capsule
     * @return the new Capsule instance.
     */
    public static Capsule create(Class cl, Xdr xdr)
	throws Exception
    {
	Capsule c;
	
	Assert.require(cl);
	
	c = (Capsule)(cl.newInstance());
	c.xdr = xdr;
	c.decode();
	return c;
    }
}
