/* 
 * CapsuleClass.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import ants.util.MD5;

/** 
 * This class (along with CodeGroup) represents all the information
 * about a type of capsule.  Both are used only
 * within ANTS.  Protocol uses CapsuleClasses & CodeGroups during
 * protocol registration.  CapsuleChannels use `em during capsule
 * execution/dispatch.  The DL* classes use `em during dynamic code
 * loading over the net.
 * <p>
 * A class can be loaded from two places:
 * <ul>
 *     <li> by the CapsuleClass(java.lang.String) constructor which
 *     loads the class from the local file system </li>
 *     <li> by a DLResponseCapsule which loads the code from another
 *     node. </li>
 * </ul>
 * <p>
 * CapsuleClass objects are only constructed in a two places:
 * <ul>
 *     <li> DLResponseCapsule
 *     <li> Protocol
 * </ul>
 * <p>
 * The fields of this class have package access since a CapsuleClass
 * may be in many different states of completion during code loading
 * and protocol registration. XXX Actually, that's probably not true.
 * 
 * @originalAuthor David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
final class CapsuleClass
{
    /**
     * The complete Class File this CapsuleClass represents.
     */
    private final byte[] code;
    
    /**
     * The name of the class this CapsuleClass represents.
     */
    private final String name;
    
    /**
     * The class this CapsuleClass represents, after its been loaded
     * of course.
     *
     * XXX make private and add accessor methods
     */
    Class cl;
    
    /**
     * Hash of the codeID + groupID + protocolID.
     * <p>
     * rename to capsuleID
     * <p>
     * XXX make private.
     *
     * @see #computeMethodID(TypeID,TypeID,TypeID)
     */
    final TypeID methodID;
    
    /**
     * The CodeGroup this CapsuleClass is a part of.
     *
     * XXX make private.
     */
    final CodeGroup group;
    
    /**
     * Get the underlying array of bytes for the Class
     * this CapsuleClass represents.
     */
    final byte[] code()
    {
	return code;
    }
    
    /**
     * Get the size of the code for the Class this CapsuleClass
     * represents.
     */
    final int codeSize()
    {
	return code.length;
    }
    
    /**
     * Get the name of this CapsuleClass.  (Also happens
     * to be the name of the Class this CapsuleClass
     * represents...
     */
    final String name()
    {
	return name;
    }
    
    /**
     * Compute a method ID based on a
     * codeID, a groupID and protocolID.
     *
     * @return the computed TypeID.
     */
    static TypeID computeMethodID(TypeID codeID,
				  TypeID groupID,
				  TypeID protocolID)
    {
	MD5 id = new MD5();
	
	// code, group, protocol is the canonical order
	id.update(codeID.unwrap());
	id.update(groupID.unwrap());
	id.update(protocolID.unwrap()); 
	// id.update(userID.getBytes())
	return TypeID.wrap(id.hash()).intern();
    }
    
    /**
     * Fetch the ProtocolID associated with this CapsuleClass.
     */
    TypeID protocolID()
    {
	Assert.assert(group);
	return group.protocolID;
    }
    
    private final UserID tmp_userID = new UserID("everyone");
    
    UserID userID()
    {
	// XXX return group.userID()
	return tmp_userID;
    }
    
    /**
     * Check that <code>this.methodID</code> matches
     * the computed method ID.  Called as a security check after
     * downloading a codegroup, to make sure that the code really is
     * part of the protocol that it claims to belong to.  This
     * should not be called for locally defined protocols (i.e., builtin).
     *
     * @return true if they match
     * @see #computeMethodID(TypeID,TypeID,TypeID)
     */
    boolean checkMethodID() {
	Assert.assert(!group.builtin());
	TypeID codeID = TypeID.wrap(MD5.hash(code));
	return methodID.equals(computeMethodID(codeID,
					       group.groupID, group.protocolID));
    }
    
    public String toString()
    {
	return "CapsuleClass[" + name
	    + "; mid=" + methodID
	    + "]";
    }
    
    /**
     * Construct a CapsuleClass with the given group, mid, name,
     * and code.  The code may be an empty array of the right
     * length which can be filled in later.
     *
     * @exception DuplicateCapsuleClassException if a capsule class is already registered under this name.
     */
    CapsuleClass(CodeGroup group, TypeID methodID, String name,
		 byte[] code)
	throws DuplicateCapsuleClassException
    {
	// until we control the local name under which 
	// a method is loaded, we need to save this info
	this.name = name;
	this.group = group;
	this.methodID = methodID;
	this.code = code;
	
	// May throw DuplicateCapsuleClassException:
	Node.currentNode().publishCapsuleClass(this);
    }
}
