/*
 * ConsoleCommand.Java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */

package ants.core;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * A base class for any commands available in the console.
 * <p>
 * XXX make the description a member variable set via the constructor.
 *
 * @see Console
 * @see ConsoleWorker
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
abstract class ConsoleCommand
{
    /**
     * The name of this command.  Set when the command
     * is added to the the table.
     *
     * @see #addCommand(String,ConsoleCommand)
     */
    private String name; 
    
    static class Alias
    {
	private final String name;
	private final ConsoleCommand cc;
	
	Alias(String name, ConsoleCommand cc)
	{
	    this.name = name;
	    this.cc = cc;
	}
	
	String getName()
	{
	    return name;
	}
	
	ConsoleCommand getCommand()
	{
	    return this.cc;
	}
    }
    
    /**
     * The table of available commands
     */
    static Hashtable commands = new Hashtable();
    
    /**
     * Return a description of the command.
     */
    abstract String getDescription();
    
    /**
     * Return the primary name of this command.
     */
    String getName()
    {
	return name;
    }
    
    /**
     * Returns additional help based on the given arguments.
     * <p>
     * args[0] - The help command name
     * args[1] - This command name
     * args[2+] - Arguments that should be described in the context of
     * this command
     */
    protected void getAdditionalHelp(ConsoleWorker cw, String args[])
    {
	cw.getOutputStream().
	    println("No additional help for this command.");
    }
    
    /**
     * Handle the command event with the given arguments
     * <p>
     * args[0] - This command name
     * args[1+] - Arguments to the command
     */
    protected abstract boolean handle(ConsoleWorker cw, String args[]);
    
    /**
     * Add a command indexed by the given name to the table of available
     * commands.
     */
    static void addCommand(String name, ConsoleCommand cc)
    {
	cc.name = name;
	ConsoleCommand.commands.put(cc.name, cc);
    }
    
    /**
     * Add an alias for an existing console command
     */
    static void addCommandAlias(String name, ConsoleCommand cc)
    {
	ConsoleCommand.commands.put(name, new Alias(name, cc));
    }
    
    /**
     * Return the collection of all available commands
     *
     * XXX returns a java.util class.
     */
    static Enumeration getCommands() 
    {
	return ConsoleCommand.commands.elements();
    }
    
    /**
     * Find the ConsoleCommand object corresponding to the given name
     */
    static ConsoleCommand findCommand(String name)
    {
	ConsoleCommand retval = null;
	Object obj;
	
	obj = ConsoleCommand.commands.get(name);
	if( obj instanceof Alias )
	    retval = ((Alias)obj).getCommand();
	else
	    retval = (ConsoleCommand)obj;
	return retval;
    }
    
    /**
     * Handle the command line specified by args.
     * <p>
     * args[0] - The name of the command
     * args[1+] - Additional arguments to the command
     */
    static boolean handleCommand(ConsoleWorker cw, String args[])
    {
	ConsoleCommand cc;
	boolean retval;
	
	cc = ConsoleCommand.findCommand(args[0]);
	if( cc != null )
	{
	    retval = cc.handle(cw, args);
	}
	else
	{
	    cw.getOutputStream().println("Command "
					 + args[0]
					 + " is not recognized.");
	    cw.getOutputStream().println("Type `help' for a list "
					 + "of valid commands");
	    retval = false;
	}
	return retval;
    }
    
    public String toString()
    {
	return "ConsoleCommand[name="
	    + this.getName()
    + "; description="
    + this.getDescription()
    + "]";
    }
}
