/* 
 * PendingRequest.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */

package ants.core;

/**
 * Used to track request capsules that need a response.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class PendingRequest
    extends NodeCache.EventHandler
{
    /**
     * The request capsule
     */
    private Capsule cap;
    
    /**
     * Amount of time to wait before retrying
     */
    private static final int defaultRetryTime = 2;
    
    /**
     * The number of times to retry
     */
    private static final int defaultRetryCount = 5;
    
    /**
     * The ttl to reprime the capsule with, this increases with each
     * retries to make sure we can get the thing out.
     */
    private int ttl = 1;
    
    /**
     * The number of times left to retry sending the capsule.
     */
    private int retries = defaultRetryCount;
    
    /**
     * Construct a pending request for this capsule
     */
    public PendingRequest(Capsule cap)
    {
	this.cap = cap;
    }
    
    /**
     * Return the request capsule
     */
    public Capsule getCapsule()
    {
	return this.cap;
    }
    
    /**
     * Event handler called by NodeCache when we're being collected.
     * This just checks if we have any retries left and sends the capsule
     * if we do.
     */
    public int collected()
    {
	int retval;
	
	if( this.retries == 0 )
	{
	    Trace.trace(Trace.ANTS_OUTGOING, "Request capsule wasn't received " +this.cap);
	    retval = 0;
	    // XXX shutdown?
	    Node.currentNode().shutdown();
	}
	else
	{
	    Node node = Node.currentNode();
	    
	    Trace.trace(Trace.ANTS_OUTGOING, "Resending capsule " +this.cap);
	    this.cap.prime(node.getAddress(), this.ttl);
	    node.routeForNode(this.cap, this.cap.getDst());
	    retval = defaultRetryTime;
	    this.retries--;
	    this.ttl++;
	}
	return retval;
    }
}
