/* 
 * PrimordialNode.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * Copyright (c) 2000, 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import ants.core.security.SecurityDefaults;
import ants.core.security.Principal;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

import edu.utah.janos.route.RouteEntry;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFormatException;
import edu.utah.janos.nodeos.Credentials;
import edu.utah.janos.nodeos.CommSpaceHandle;
import edu.utah.janos.nodeos.CommSpaceElement;
import edu.utah.janos.nodeos.CommQueuePusher;
import edu.utah.janos.nodeos.CommQueuePuller;
import edu.utah.janos.nodeos.Flow;
import edu.utah.janos.nodeos.FlowStartSpec;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.MemSpec;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.OutChannel;
import edu.utah.janos.nodeos.ThreadPoolSpec;

import edu.utah.janos.io.Log;
import edu.utah.janos.io.LogListener;
import edu.utah.janos.io.DatagramLogListener;

import java.io.IOException; // XXX

import java.util.Hashtable; // XXX
import java.util.Enumeration; // XXX
import java.util.Vector; // XXX


/**
 * The root Node object that boots the system and tracks system-wide
 * state.  This is as opposed to the per-flow Node interface
 * that every protocol/application sees.
 * <p>
 * XXX move the ConfigurationManager stuff into here.
 * <p>
 * XXX Move the command line option stuff OUT of here.  Make start()
 * method take functions, or make them static method (e.g, if can be
 * changed at run-time).
 * <p>
 * XXX Should be called the "Primordial Flow"
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class PrimordialNode
{
    static 
    {
	System.out.println("ANTS v2.0");
    }
    
    /**
     * Array of package names that are considered part of the ANTS system.
     */
    static final String antsPackages[] = {
	"ants.core",
	"ants.core.security",
	"ants.util",
    };
    
    /**
     * Array of classes that should be exported to downloaded code. Note
     * that ANTS applications are not subject to this restricted list
     * and may access any visible class.  We really should provide (looser)
     * restrictions on applications, though.
     * <p>
     * Note that only the Exceptions in the <code>ants.core.security</code> 
     * package are visible to downloaded code.  This is required.
     */
    static final String exportedAntsClasses[] = {
	/* ants.core.* */
	"ants.core.Application",
	"ants.core.Assert",
	"ants.core.Entity",
	"ants.core.Extension",
	"ants.core.AttachException",
	"ants.core.ByteArray",
	"ants.core.ConfigurationManager",
	"ants.core.DataCapsule",
	"ants.core.DataProtocol",
	"ants.core.DuplicateCapsuleClassException",
	"ants.core.FullCodeGroupException",
	"ants.core.NeighborTable",
	"ants.core.NoSuchApplicationError",
	"ants.core.Node",
	"ants.core.NodeAddress",
	"ants.core.NodeCache",
	"ants.core.NodeConfigure",
	"ants.core.NodeHasNoAddressError",
	"ants.core.NodeState",
	"ants.core.Protocol",
	"ants.core.ResourceLimitException",
	"ants.core.Capsule",
	"ants.core.TypeID",
	"ants.core.Xdr",

	/* ants.core.security */
	"ants.core.security.ReferenceMonitorException",
	
	/* ants.util.* */
	"ants.util.ExtraKeyArgsError",
	"ants.util.InvalidKeyArgsException",
	"ants.util.KeyArgs",
	"ants.util.LineListException",
	"ants.util.LineListTokenizer",

	/*
	 * Several miscellaneous JNodeOS classes that *protocols*
	 * should be able to see.
	 */
	"edu.utah.janos.route.RouteEntry",
	"edu.utah.janos.nodeos.Address",
	"edu.utah.janos.nodeos.AddressFactory",
	"edu.utah.janos.nodeos.AddressFormatException",
	"edu.utah.janos.nodeos.NoSuchNodeAddress",
    };
    
    /**
     * Arguments the PrimordialNode understands and their defaults.
     */
    private final static String[] defaults = {"-consoleport",  null,
					      "-msgprefix",    null,
					      "-onetrick",     null,
					      "-log",          "0",
					      "-routes",       null,
					      "-queryport",    null};
    
    /**
     * Port to start a console thread on this port (if > -1).
     */
    private static int consolePort  = -1;
    private static int watchDogPort = -1;
    private static int queryPort    = -1;
    
    private static long startupTime = System.currentTimeMillis();
    
    /**
     * Is this a one-trick node?
     *
     * @see #oneTrick()
     */
    private static boolean oneTrickPony = false;
    
    private final static Log logger = new Log("sys");
    
    /**
     * XXX currently node-level logging just goes
     * to stdout.
     */
    static class StdoutListener
	implements LogListener
    {
	public void logMsg(int level, String msg)
	{
	    if (level <= Log.WARNING)
		System.out.println(msg);
	}
    }
    
    static
    {
	logger.addListener(new StdoutListener());
    }
    
    private static String msgPrefix = null;
    
    private static String policyFile = null;
    
    /**
     * The name of the file to read initial route information
     * from.  Defaults to null (no default routes).
     */
    private static String routefile = null;
    private static boolean updateRoutes = false;
    private static RouteTable routeTable;
    private static NeighborTable neighborTable;
    
    /**
     * 32-bit logical address of this Node.
     */
    private static int address;

    /**
     * @see #setPhysicalAddress(String,String)
     */
    private static String protocolSpec;

    /**
     * @see #setPhysicalAddress(String,String)
     */
    private static String addressSpec;

    /**
     * @see #setProxyAddress(String,String)
     */
    private static String proxyProtocolSpec;

    /**
     * @see #setProxyAddress(String,String)
     */
    private static String proxyAddressSpec;
    
    // Maps process IDs to CommQueues:
    private static final Hashtable pidToFlowQueue = new Hashtable();
    
    // Maps application ports to their CommQueues
    private static final Hashtable portToAppQueue = new Hashtable();
    
    /**
     * Global cache of packets for MIDs that haven't been loaded/resolved.
     * <p>
     * The Dead Letter Office.
     */
    static final UnknownMidPackets unknownMidPackets =
	new UnknownMidPackets();
    
    /**
     * Global object used for gathering statistics.
     */
    static final NodeStat stats = new NodeStat();
    
    /**
     * No instance of this class is ever created.
     */
    private PrimordialNode()
    {
    }
    
    static CommQueuePusher getProtocolQueue(TypeID pid)
    {
	return (CommQueuePusher)pidToFlowQueue.get(pid);
    }
    
    static CommQueuePusher getAppQueue(int port)
    {
	return (CommQueuePusher)portToAppQueue.get(new Integer(port));
    }
    
    static long getStartupTime()
    {
	return startupTime;
    }
    
    /**
     * @return The NodeStat object that holds all of the stats for this
     * node.
     */
    static NodeStat getStats()
    {
	return stats;
    }
    
    /**
     * Start the node on its way.
     */
    static void start(Vector appConfigs, Vector extConfigs)
	throws IOException, AddressFormatException, NoSuchNodeAddress,
	InterruptedException, ClassNotFoundException
    {
	int lpc;
	
	Assert.assert(appConfigs);
	Assert.assert(extConfigs);
	
	/*
	 * Define the local packages for ANTS.  These are
	 * packages that will be loaded into the root (e.g,
	 * system) classloader.  All others will be loaded on
	 * a per-flow basis.  Also, seal the packages so that
	 * noone can try to add to them.
	 *
	 * Look at the notes in core.FlowClassLoader.
	 */
	for( lpc = 0; lpc < antsPackages.length; lpc++ )
	{
	    Flow.currentFlow().addSystemPackage(antsPackages[lpc]);
	    Flow.currentFlow().sealPackage(antsPackages[lpc]);
	}
	
	/* Startup the primordial message handler */
	Flow.currentFlow().threadStart(new PrimordialMessenger(),
				       "Primordial Messenger");
	
	/*
	 * Load the node's security policy.  Null implies the
	 * default policy in SecurityDefaults.
	 */
	ants.core.security.ReferenceMonitor.loadPolicy(policyFile);
	
	/*
	 * Create a Node interface for the builtin protocols.
	 */
	Node node = new Node(Flow.currentFlow(),
			     unknownMidPackets,
			     SecurityDefaults.ADMIN);
	
	/*
	 * Setup the route table and neighbor table.  Initialize both
	 * from the routefile argument.  The neighbor table has a
	 * run-time handler for notification on updates.
	 */
	neighborTable = new NeighborTable();
	neighborTable.startHandler(node);

	routeTable = new RouteTable();
	RouteTable.initializeTables(getAddress(), routefile,
				    routeTable, neighborTable);
	
	routefile = null; // drop the route file name
	
	Assert.assert(getAddress() != 0);
	
	/*
	 * XXX Perhaps the protocols started in here should be created
	 * inside their own flows?  Or do they need access to some
	 * global state?
	 */
	try
	{
	    /*
	     * Create the DLProtocol.  
	     *
	     * Manually associate it with a key that matches
	     * just the ANTS header.  This way it will get all
	     * packets that aren't matched by more specific
	     * filters.
	     *
	     * XXX runs with the same ThreadPool as other root flow tasks...
	     * ... might be okay actually (nothing else for those threads).
	     */
	    DLProtocol dlp = new DLProtocol();
	    
	    Trace.trace(Trace.ANTS_STARTUP, "Installing " +dlp);
	    
	    /*
	     * Activate the MIDIDRequest/Response parts of
	     * the DLProtocol.  (The other parts
	     * aren't activated because they're inert --
	     * see BuiltinProtocol.)
	     */
	    node.register(dlp);
	    
	    /*
	     * Setup the DataProtocol.  Self-forwarding data
	     * packet service is provided by the Node.
	     *
	     * XXX this should be tuned & optimized.
	     */
	    DataProtocol dp = new DataProtocol();
	    Trace.trace(Trace.ANTS_STARTUP, "Installing " +dp);
	    node.register(dp);
	    
	    // XXX unknown ANEP packet forwarder--- a cut-through channel!.
	}
	catch (Exception e)
	{
	    Node.abort(e, "Builtin protocol setup failure.");
	}
	
	/*
	 * Start the dynamic routing (if necessary).
	 *
	 * XXX an option is needed to override 'ants.route.DynamicRouteApp'.
	 *
	 * XXX deprecate updateRoutes. 
	 */
	if (updateRoutes)
	{
	    // Start the app that maintains the routing table
	    AppConfig dynrouteApp;
	    
	    dynrouteApp = new AppConfig("ants.route.DynamicRouteApp",
					new String[] {
						"-principal", "RouteUser"
					}, 0);
	    fireUpNewAppFlow(dynrouteApp);
	}
	
	/*
	 * Start up the applications on this node.
	 */
	Enumeration apps = appConfigs.elements();
	while(apps.hasMoreElements())
	{
	    AppConfig a = (AppConfig)apps.nextElement();
	    fireUpNewAppFlow(a);
	}
	
	
	/*
	 * XXX MIT ANTS supports Node extensions and installs
	 * them at this point.  Since they're unused, we ignore
	 * them for now.
	 */
	Assert.assert(extConfigs.size() == 0);
	
	
	/*
	 * Start the console.
	 */
	//startConsole(node, consolePort);
	//startWatchDog(node, watchDogPort);
	if( queryPort > -1 )
	{
	    DatagramLogListener dll =
		new DatagramLogListener();
	    
	    try
	    {
		dll.start(queryPort);
		logger.addListener(dll);
	    }
	    catch(Exception e)
	    {
	    }
	}
    }
    
    
    private static void startConsole(Node node, int consolePort)
    {
	if (consolePort > -1)
	{
	    Console con = null;
	    int tries = 1;
	    
	    while ((tries != 0) && (tries < 15))
	    {
		try
		{
		    con = new Console(node, consolePort);
		    tries = 0;
		}
		catch (IOException ioe)
		{
		    tries++;
		    consolePort++;
		}
	    }
	    
	    if (con != null)
	    {
		Flow.currentFlow().threadStart(con, "Root Console");
		node.log(Log.NOTICE,
			 "Started a control console on "
			 +node.hostname()+ ":" +con.getPort());
	    }
	    else
	    {
		node.log(Log.NOTICE,
			 "Unable to start a control console on port " +
			 con.getPort());
	    }
	}
    }
    
    private static void startWatchDog(Node node, int watchDogPort)
    {
	if(watchDogPort > -1)
	{
	    WatchDog wd = null;
	    int tries = 1;
	    
	    while( (tries != 0) && (tries < 15) )
	    {
		try
		{
		    wd = new WatchDog(node, watchDogPort);
		    tries = 0;
		}
		catch(IOException e)
		{
		    tries++;
		}
	    }
	    
	    if( wd != null )
	    {
		Flow.currentFlow().threadStart(wd, "Root Watchdog");
		node.log(Log.NOTICE,
			 "Started a watch dog on port "
			 +node.hostname()+ ":" +wd.getPort());
		
	    }
	    else
	    {
		node.log(Log.NOTICE,
			 "Unable to start a watch dog on port "
			 + watchDogPort);
	    }
	}
    }
    
    static NeighborTable getNeighborTable () {
	return neighborTable;
    }
    

    private static boolean gaveOutRouteTable = false;

    /**
     * Return the global route table.  This provides
     * both read and write access to the RouteTable.
     * Because of the lack of synchronization, we only
     * allow a single app to get access to the route
     * table at this level.
     */
    static RouteTable getRouteTable()
    {
	if (gaveOutRouteTable)
	    throw new Error("Access to the global route table was already given out.");

	gaveOutRouteTable = true;
	return routeTable;
    }
    
    /**
     * Return the logical ANTS address of the next node
     * on the path to the given addr.
     */
    static int nextHopToNode(int addr)
    {
	// Assert.assert(addr != ..);
	RouteEntry re = routeTable.getRoute(addr);
	if (re == null)
	{
	    Trace.trace(Trace.ANTS_ROUTEINFO, "Couldn't get route for " +Address.toString(addr));
	    return 0;
	}
	
	return re.getLogicalNext();
    }
    
    /**
     * Return the physical address of the next node on the path to
     * the given addr.  Returns null if no route is available.
     *
     * @param addr the logical address of the ultimate destination
     * @return the Address representing that node
     */
    static Address nextPhysicalToNode(int addr)
    {		
	RouteEntry re = routeTable.getRoute(addr);
	if (re == null)
	{
	    Trace.trace(Trace.ANTS_ROUTEINFO, "Couldn't get route for " +Address.toString(addr));
	    return null;
	}
	
	return re.getPhysicalNext();
    }
    
    /**
     * Return the physical address of the given neighbor node.
     * The given address should be a logical address.
     *
     * @param neighbor a neighbor node's logical address
     */
    static Address nextPhysicalToNeighbor(int neighbor) {
	Address a = neighborTable.getNeighbor(neighbor);
	
	if (a == null)
	{
	    Trace.trace(Trace.ANTS_ROUTEINFO, "Couldn't get link for " +Address.toString(neighbor));
	    return null;
	}
	return a;
    }
    
    
    /*
     * Parse a set of KeyArgs arguments.  Adds '-routes
     * <filename>', '-log <int>', '-policy <filename>' and
     * '-updateRoutes' arguments.  Will throw an
     * InvalidKeyArgsException if any arguments are not understood
     * by Node or its superclases. 
     * <p>
     * Defaults '-updateRoutes' to false.  Defaults route
     * filename to null.  Defaults policy file to null (none).
     *
     * @see #defaults
     */
    static void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	
	for (int i = 0; i < k.length(); i++) 
	{
	    if (k.key(i).equals("-routes")) 
	    {
		routefile = k.arg(i);
		k.strike(i);
	    }
	    else if (k.key(i).equals("-msgprefix"))
	    {
		msgPrefix = k.arg(i);
		logger.setPrefix(msgPrefix);
		k.strike(i);
	    }
	    else if (k.key(i).equals("-onetrick"))
	    {
		oneTrickPony = true;
		k.strike(i);
	    }
	    else if (k.key(i).equals("-consoleport"))
	    {
		// Trace.trace(Trace.ANTS_STARTUP, "Parsing consoleport: " +k.arg(i)+ ".");
		if (k.arg(i) != null)
		    consolePort = Integer.parseInt(k.arg(i));
		k.strike(i);
	    }
	    else if (k.key(i).equals("-updateRoutes"))
	    {
		// updateRoutes = Boolean.valueOf(k.arg(i)).booleanValue();
		updateRoutes = true;
		k.strike(i);
	    }
	    else if (k.key(i).equals("-log")) 
	    {
		// XXX ignore -log for now
		k.strike(i);
	    }
	    else if (k.key(i).equals("-watchdog"))
	    {
		if (k.arg(i) != null)
		    watchDogPort = Integer.parseInt(k.arg(i));
		k.strike(i);
	    }
	    else if (k.key(i).equals("-queryport"))
	    {
		if (k.arg(i) != null)
		    queryPort = Integer.parseInt(k.arg(i));
		k.strike(i);
	    }
	    else if (k.key(i).equals("-policy")) {
		policyFile = k.arg(i);
		k.strike(i);
	    }
	}
	
	k.noMore();
    }
    
    static boolean oneTrick()
    {
	return oneTrickPony;
    }
    
    /**
     * Shut down the entire VM.  Only invoked (occasionally)
     * by Node.shutdown().
     *
     * @see ants.core.Node#shutdown()
     */
    static void shutdown()
    {
	edu.utah.janos.nodeos.Node.shutdown();
    }
    
    /**
     * Return the logical (ANTS-level) address of this node as
     * a 32-bit integer.  If the address is zero, that means
     * it has not be set yet.  (It can be set at run-time by
     * DANTE or another dynamic address specification protocol.)
     * <p>
     * It is the caller's responsibility to detect an unset address
     * and wait or throw an exception.
     */
    static int getAddress()
    {
	return PrimordialNode.address; // XXX
    }
    
    /**
     * Set this node's logical address.  
     * XXX assert given address is not zero?
     *
     * @see ConfigurationManager#parseCommands(Vector,String[])
     */
    static void setAddress(int address)
    {
	PrimordialNode.address = address;
    }	
    
    /**
     * Set the physical address of this node.  The physical
     * address is the address at which this node will receive
     * packets.  It is used by the CapsuleChannel code to
     * create input channels on the appropriate interface.
     * <p>
     * The specs are not parsed, they're just blindly passed
     * to the Java NodeOS.
     *
     * @param protocolSpec the String representation of the protocol stack to use.
     * @param addressSpec the String representation of this node's physical address.
     * @see #setProxyAddress(String,String)
     */
    static void setPhysicalAddress(String protocolSpec, String addressSpec)
    {
	Assert.assert(PrimordialNode.protocolSpec == null);
	Assert.assert(PrimordialNode.addressSpec == null);
	PrimordialNode.protocolSpec = protocolSpec;
	PrimordialNode.addressSpec = addressSpec;
    }
    
    /** 
     * The proxy's receive address, for nodes that are being proxied.
     * Normally this is null (no proxy).  If it is not null, it is the
     * address to which remote nodes should send packets (i.e., the
     * proxy).  In contrast, the physical address is where this node
     * should receive packets from (i.e., after the proxy hands them off).
     *
     * @param protocolSpec the String representation of the protocol stack to send to the proxy.
     * @param addressSpec the String representation of the proxy's physical address.
     * @see #setPhysicalAddress(String,String)
     */
    static void setProxyAddress(String protocolSpec, String addressSpec)
    {
	PrimordialNode.proxyProtocolSpec = protocolSpec;
	PrimordialNode.proxyAddressSpec = addressSpec;
    }

    /**
     * @see #setPhysicalAddress(String,String)
     */
    static String getReceiveProtocolSpec()
    {
	Assert.assert(PrimordialNode.protocolSpec != null);
	return PrimordialNode.protocolSpec;
    }

    /**
     * @see #setPhysicalAddress(String,String)
     */
    static String getReceiveAddressSpec()
    {
	Assert.assert(PrimordialNode.addressSpec != null);
	return PrimordialNode.addressSpec;
    }

    /**
     * @see #setProxyAddress(String,String)
     * @see #setPhysicalAddress(String,String)
     */
    static String getVisibleProtocolSpec()
    {
	Assert.assert(PrimordialNode.protocolSpec != null);
	if (PrimordialNode.proxyProtocolSpec != null)
	    return PrimordialNode.proxyProtocolSpec;
	return PrimordialNode.protocolSpec;
    }

    /**
     * @see #setProxyAddress(String,String)
     * @see #setPhysicalAddress(String,String)
     */
    static String getVisibleAddressSpec()
    {
	Assert.assert(PrimordialNode.addressSpec != null);
	if (PrimordialNode.proxyAddressSpec != null)
	    return PrimordialNode.proxyAddressSpec;
	return PrimordialNode.addressSpec;
    }
    
    static void message(String msg)
    {
	logger.logMsg(Log.EMERG, msg);
    }
    
    static void message(int level, String msg)
    {
	logger.logMsg(level, msg);
    }
    
    static void dumpStackTrace(int level, Throwable exc)
    {
	logger.logThrowable(level, exc);
    }
    
    static void detachApplication(int port)
    {
	portToAppQueue.remove(new Integer(port));
    }
    
    /**
     * Add the given MID to the set understood by this node.
     * This creates a flow to handle the mid (using the
     * resource defined by the userInfo field).
     * <p>
     * XXX Currently the whole method is synchronized.  This
     * protects the race-prone lookup/create in the pidToFlowQueue
     * map.  It doesn't need to protect the fireUpNewProtocolFlow.
     *
     * @param pid the PID of the new flow.
     * @param userInfo the user identity/credentials of the new flow
     * @param mid the MID that caused the flow to get loaded.
     * @param prevNode the node that sent the initial MID-labeled packet.
     */
    static void addMID(TypeID pid,
		       UserID userInfo,
		       TypeID mid,
		       TypeID gid,
		       int prevNode)
    {
	CommQueuePusher queue = PrimordialNode.getProtocolQueue(pid);
	
	if(queue == null)
	{
	    Trace.trace(Trace.ANTS_DYNLOAD, "addMID(pid=" + pid +") is new.");
	    
	    // Add new pid to the Protocol Register
	    ProtocolRegister.Owner owner;
	    owner = new ProtocolRegister.Owner(pid.toString(), "Downloaded");
	    ProtocolRegister.register(pid, owner);
	    
	    // Start a new flow for the new pid
	    queue = new CommQueuePusher();
	    try
	    {
		fireUpNewProtocolFlow(pid, queue);
		pidToFlowQueue.put(pid, queue);
	    }
	    catch (Throwable th)
	    {
		Node.abort(th, "Uh-oh.  fireUpNewProtocolFlow() blew up.");
	    }
	}
	else
	{
	    Trace.trace(Trace.ANTS_DYNLOAD, "addMID(pid=" + pid +") Q already setup...");
	}
	
	/*
	 * Even if the pid is already registered, the MID
	 * probably isn't.  Push MID info into the queue for
	 * the new flow to snack on.
	 */
	CommSpaceElement msg[] = {
	    new CommSpaceElement(ProtocolBooster.NEW_MID),
	    new CommSpaceElement(mid.unwrap()),
	    new CommSpaceElement(gid.unwrap()),
	    new CommSpaceElement(prevNode)
	};
	queue.enqueue(msg);
    }
    
    /**
     * Install a protocol, this will start up the a new flow if necessary
     * and load in the code.
     *
     * @param pid The protocol ID to be loaded
     * @param name The class name of the protocol to install
     */
    static void installProtocol(TypeID pid, String name, Principal principal)
    {
	CommQueuePusher queue = PrimordialNode.getProtocolQueue(pid);
	
	/* Find the current owner of the protocol if there is one */
	ProtocolRegister.register(pid,
				  new ProtocolRegister.Owner(name, "Local"));
	if( queue == null )
	{
	    /*
	     * We're the owner now, make a new queue and start up a
	     * new flow
	     */
	    queue = new CommQueuePusher();
	    try
	    {
		fireUpNewProtocolFlow(pid, queue);
		pidToFlowQueue.put(pid, queue);
	    }
	    catch(Throwable th)
	    {
		Node.abort(th, "Uh-oh.  fireUpNewProtocolFlow() blew up.");
	    }
	}
	CommSpaceElement msg[] = {
	    new CommSpaceElement(ProtocolBooster.LOCAL_PROTOCOL),
	    new CommSpaceElement(name),
	    new CommSpaceElement(principal.getName())
	};
	queue.enqueue(msg);
    }
    
    // XXX Move the following functions into a FlowManager class.
    
    static void fireUpNewAppFlow(AppConfig app)
	throws InterruptedException,
	IOException
    {
	CommQueuePusher pusher = new CommQueuePusher();
	BoosterArgs ba = fireUpNewFlow(new Credentials("boot app"), // XXX
				       new ThreadPoolSpec(8, 0, null), // XXX
				       new MemSpec(), // Does an application need UdpBuffers?
				       null,
				       pusher,
				       AppBooster.class.getName(),
				       app.args());
	
	/* Negotiate the port number for the application */
	int port = ba.waitForAppPort();
	Integer key;
	
	if( port == 0 )
	{
	    /* They'll take any port number */
	    boolean done = false;
	    
	    port = 128;
	    while( (port > 0) && !done )
	    {
		key = new Integer(port);
		
		if( portToAppQueue.containsKey(key) )
		{
		    port--;
		}
		else
		{
		    portToAppQueue.put(key, pusher);
		    done = true;
		}
	    }
	}
	else
	{
	    key = new Integer(port);
	    
	    /* Check if someone already took the port */
	    if( portToAppQueue.containsKey(key) )
	    {
		port = 0;
	    }
	    else
	    {
		/*
		 * Add them to the port
		 * XXX This isn't atomic
		 */
		portToAppQueue.put(key, pusher);
	    }
	}
	/*
	 * Reply back to the app, if port == 0 then the negotiation
	 * failed.
	 */
	pusher.enqueue(new CommSpaceElement(port));
    }
    
    static void fireUpNewProtocolFlow(TypeID pid, CommQueuePusher queue)
	throws InterruptedException,
	IOException
    {
	String[] args = new String[0];
	
	fireUpNewFlow(new Credentials("dlprotocol:" +pid), // XXX
		      new ThreadPoolSpec(8, 0, null), // XXX
		      new MemSpec(), // XXX
		      pid,
		      queue,
		      ProtocolBooster.class.getName(),
		      args);
    }
    
    static BoosterArgs fireUpNewFlow(Credentials cred,
				     ThreadPoolSpec tps,
				     MemSpec ms,
				     TypeID pid,
				     CommQueuePusher queue,
				     String className,
				     String[] args)
	throws InterruptedException,
	IOException
    {
	Assert.require(className);
	Assert.require(args);
	
	/* Setup the arguments for our new flow */
	BoosterArgs ba = new BoosterArgs(pid,
					 unknownMidPackets.midToSpaceMap);
	
	FlowStartSpec fss = new FlowStartSpec(className,
					      args,
					      ba.getCommSpaceHandle());
	
	Runnable th = new Runnable() {
	    public void run() {
		/* XXX do cleanup for the flow. */
	    }
	};
	
	Flow.currentFlow().newFlow(cred, tps, ms, fss, th);
	
	/*
	 * Bind our local queue object to the one passed through
	 * the arguments space.
	 */
	queue.bind(ba.waitForFlowMsgQueue());
	return ba;
    }
}
