/* 
 * Permission.java
 *
 * Copyright (c) 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core.security;

/**
 * Describes an abstract permission, consisting of a string name and
 * an optional set of actions from the set {read,write,execute}.
 * 
 * @author Andrew Whitaker
 */
public class Permission
    implements java.io.Serializable
{
    protected static final int READ = 0x01;
    protected static final int WRITE = 0x02;
    protected static final int EXECUTE = 0x04;
    
    protected int actionMask;
    protected String name;
    
    /**
     * Creates a new permission with the given human-readable name and
     * no actions.  Names serve as unique keys.
     */
    public Permission(String name)
    { 
	this.name = name;
	actionMask = 0;
    }
    
    /**
     * Creates a permission with a set of actions.  Actions are
     * specified by passing an argument string consisting of a subset of
     * {"read","write","execute"}.  The actions must be lower-case.  For
     * example: new Permission ("Routing table","read write")
     */
    public Permission(String name,String actions) 
    { 
	this(name);
	
	actionMask = 0;
	if (actions.indexOf("read") != -1)
	    actionMask |= READ;
	if (actions.indexOf("write") != -1)
	    actionMask |= WRITE;
	if (actions.indexOf("execute") != -1)
	    actionMask |= EXECUTE;
	
    }
    
    /**
     * Returns the actions as a human-readable string 
     */
    public String getAction() 
    {
	StringBuffer action = new StringBuffer("");
	
	if ( (actionMask | READ) == actionMask)
	    action.append("read ");
	if ( (actionMask | WRITE) == actionMask)
	    action.append("write ");
	if ( (actionMask | EXECUTE) == actionMask)
	    action.append("execute");
	
	return action.toString().trim();
    }
    
    public String getName()
    {
	return name;
    }
    
    /**
     * test whether this permission implies the given permission.  For
     * example,  read/write implies the read permission.  Names are
     * case-insensitive.  
     */
    
    public boolean implies(Permission p)
    {
	return ( (p.getName().equalsIgnoreCase(this.getName())) &&
		 ( (p.actionMask | this.actionMask) == this.actionMask)   );
    }
    
    public String toString()
    {
	return (this.getName() + " " + this.getAction());
    }
    
    /**
     * Tests for exact equality: the name and actions must precisely
     * match.  Names are case-insensitive.
     */
    public boolean equals(Object o)
    {
	if (this == o)
	    return true;
	else if (! (o instanceof Permission))
	    return false;
	else
	{
	    Permission np = (Permission) o;
	    return (  (np.getName().equalsIgnoreCase(this.getName()))
		      && (np.actionMask == this.actionMask));
	}
    }
}


