/* 
 * ReferenceMonitor.java
 *
 * Copyright (c) 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core.security;

import java.io.IOException;
import java.io.FileInputStream;
import java.io.ObjectInputStream;


/**
 * This class serves as the interface to a basic reference monitor.
 * Ants uses this class to check access to sensitive data and
 * operations (i.e., the routing table; system shutdown; etc.)
 * 
 * Note that this class supplies mechanism, not policy.  The security
 * policy is specified by loading a serialized instance of a Policy
 * object.  The policy object is specified by using -policy flag to
 * the node argument in the ANTS configuration file:
 *
 * node 1.1.2.5 -updateRoutes -policy ~/ants/policy.ser
 *
 * If no policy object is specified, the ReferenceMonitor loads a
 * default policy from the SecurityConstants class.  If there is an
 * error loading the policy file, the ReferenceMonitor will fail fast
 * rather than passively loading the default policy.
 *
 * @author Andrew Whitaker
 * @see Policy
 * @see BasicPolicy
 * @see SecurityDefaults
 */
public final class ReferenceMonitor
{
    private static Policy policy = null;
    
    /**
     * Sets the access control policy at startup.  The argument can be
     * either a file containing a serialized policy object; or null
     * which implies the default policy in SecurityDefaults.  For
     * security reasons, the policy can only be set once.  If there is
     * an error loading the policy file, this method will throw an
     * exception rather than passively load the default policy.
     */
    public static void loadPolicy (String file)
	throws IOException,ClassNotFoundException
    {
	if (policy != null)
	    return;
	else if (file == null)
	    policy = SecurityDefaults.defaultPolicy;
	else 
	    policy = loadFromFile(file);
    }
    
    private static Policy loadFromFile(String file) 
	throws IOException, ClassNotFoundException
    {
	FileInputStream fis = new FileInputStream(file);
	ObjectInputStream in = new ObjectInputStream(fis);
	Policy p = (Policy) in.readObject();
	in.close();
	fis.close();
	return p;
    }
    
    /**
     * Check whether the given principal has a permission.  If the
     * policy hasn't been specifed, this method will fail.
     */
    public static void checkPermission(Principal who,Permission what) 
	throws ReferenceMonitorException
    {
	
	if ( (policy == null)
	     || (!policy.hasPermission(who,what)))
	    throw new ReferenceMonitorException("Invalid access: "
						+who.toString()+
						" " + what.toString());
    }
    
}


