/* 
 * SecurityDefaults.java
 *
 * Copyright (c) 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core.security;

import java.io.FileOutputStream;
import java.io.ObjectOutputStream;

/**
 * Represents a set of built-in access control defaults.  These
 * defaults can be overridden by loading a serialized policy object.
 *
 * The main routine of this class serializes the default policy.
 *
 * @author Andrew Whitaker
 * @see ReferenceMonitor
 */
public final class SecurityDefaults
{
    /** 
     * Any change to the name of this object must be reflected in
     * Primordial Node.
     */
    public static final Principal ADMIN
	= new Principal("Admin");
    
    /** The initial principal for all nodes */
    public static final Principal BOOTSTRAP_USER
	= new Principal("BootStrap");
    
    /** The default principal for the routing application */
    public static final Principal ROUTE_USER
	= new Principal("RouteUser");
    
    /** The principal for the dante application */
    public static final Principal DANTE_USER
	= new Principal("DanteUser");
    
    /** The default principal for local applications and protocols */
    public static final Principal LOCAL_USERS
	= new Principal("LocalUsers");
    
    /** The default principal for dynamicaly loaded protocols */
    public static final Principal REMOTE_USERS
	= new Principal("RemoteUsers");
    
    
    /** Permission to modify the routing table */
    public static final Permission MODIFY_ROUTES
	= new Permission("routing.table","read write");
    
    /** Permission to read the neighbor table */
    public static final Permission READ_NEIGHBORS
	= new Permission("neighbor.table","read");  
    
    /** Permission to modify the neighbor table */
    public static final Permission MODIFY_NEIGHBORS
	= new Permission("neighbor.table","read write");  
    
    /** Permission to shutdown the node */
    public static final Permission SHUTDOWN
	= new Permission("shutdown","execute");
    
    /** Permission to write to the local log */
    public static final Permission LOG
	= new Permission("log","execute");
    
    /** permission to set the local address of the node */
    public static final Permission SET_ADDRESS
	= new Permission("address","write");
    
    /** the "root" permission: this is dangerous */
    public static final Permission ADMIN_PERMISSION
	= new AdminPermission();
    
    /** permission to read the local node's physical connection spec */
    public static final Permission READ_PHYSICAL_PROTOCOL_SPEC
	= new Permission ("physicalProtocolSpec","read");
    
    /** permission to read the local node's physical address */
    public static final Permission READ_PHYSICAL_ADDRESS_SPEC
	= new Permission ("physicalAddressSpec","read");
    
    /** permission to use the Node's local send method */
    public static final Permission LOCAL_SEND
	= new Permission ("localSend","execute");
    
    /** permission to start up a thread */
    public static final Permission START_THREAD
	= new Permission ("threadStart","execute");
    
    /** 
     * permission to send to directly to neighbor, bypassing TTL and
     * source checks 
     */
    public static final Permission RAW_SEND
	= new Permission ("rawSend","execute");
    
    /*package*/ static final Policy defaultPolicy = new BasicPolicy();
    
    // load the default policy
    static
    {
	// the ADMIN is omnipotent
	defaultPolicy.addPermission(ADMIN,ADMIN_PERMISSION);
	
	defaultPolicy.addPermission(BOOTSTRAP_USER,LOG);
	defaultPolicy.addPermission(BOOTSTRAP_USER,START_THREAD);
	
	defaultPolicy.addPermission(ROUTE_USER,MODIFY_NEIGHBORS);
	defaultPolicy.addPermission(ROUTE_USER,MODIFY_ROUTES);
	defaultPolicy.addPermission(ROUTE_USER,LOG);
	defaultPolicy.addPermission(ROUTE_USER,LOCAL_SEND);
	defaultPolicy.addPermission(ROUTE_USER,RAW_SEND);
	defaultPolicy.addPermission(ROUTE_USER,READ_PHYSICAL_ADDRESS_SPEC);
	defaultPolicy.addPermission(ROUTE_USER,READ_PHYSICAL_PROTOCOL_SPEC);
	
	defaultPolicy.addPermission(DANTE_USER,LOCAL_SEND);
	defaultPolicy.addPermission(DANTE_USER,MODIFY_ROUTES);
	defaultPolicy.addPermission(DANTE_USER,MODIFY_NEIGHBORS);
	defaultPolicy.addPermission(DANTE_USER,SET_ADDRESS);
	defaultPolicy.addPermission(DANTE_USER,LOG);
	defaultPolicy.addPermission(DANTE_USER,RAW_SEND);
	defaultPolicy.addPermission(DANTE_USER,START_THREAD);
	defaultPolicy.addPermission(DANTE_USER,READ_PHYSICAL_ADDRESS_SPEC);
	defaultPolicy.addPermission(DANTE_USER,READ_PHYSICAL_PROTOCOL_SPEC);
	
	defaultPolicy.addPermission(LOCAL_USERS,LOG);
	defaultPolicy.addPermission(LOCAL_USERS,LOCAL_SEND);
	defaultPolicy.addPermission(LOCAL_USERS,READ_NEIGHBORS);
	defaultPolicy.addPermission(LOCAL_USERS,START_THREAD);
	
	defaultPolicy.addPermission(REMOTE_USERS,LOG);
	defaultPolicy.addPermission(REMOTE_USERS,READ_NEIGHBORS);
    }
    
    /**
     * Serializes the default policy and dumps it in a file
     */
    public static void main(String [] args)
	throws Exception
    {
	if (args.length == 0)
	{
	    System.err.println ("Use: java SecurityDefaults outputFile");
	    System.exit(0);
	}
	else
	{
	    FileOutputStream fis = new FileOutputStream(args[0]);
	    ObjectOutputStream out = new ObjectOutputStream(fis);
	    out.writeObject(defaultPolicy);
	    out.flush();
	    fis.flush();
	    out.close();
	    fis.close();
	}
    }
}
