/* 
 * DanteCapsule.java
 *
 * Copyright (c) 2000 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.dante;

import ants.core.Node;
import ants.core.Capsule;
import ants.core.Xdr;

/** 
 * Negotiates a dynamic address on behalf of an ABONE edge node.  This
 * capsule must operate at the so-called virtual-link layer because
 * tunnels and routes have not yet been established.  
 *
 * This class takes some liberties with the official DANTE header
 * specification.  For example, the core virtual link layer address
 * is not passed.
 *
 *@author Andrew Whitaker
 */
public final class DanteCapsule
    extends Capsule 
    implements DanteConstants
{  
    public byte[] mid() { return findMID("ants.dante.DanteCapsule"); }
    //  public byte[] gid() { return findGID("ants.dante.DanteCapsule");}    
    public byte[] pid() { return findPID("ants.dante.DanteCapsule");}
    
    protected int type;
    protected String edgePhysAddress;
    protected String edgeProtocol;
    protected int edgeVNA;
    protected int coreVNA;
    
    public void setEdgeVNA(int a) {
	edgeVNA  = a;
    }
    public int getEdgeVNA () {
	return edgeVNA;
    }
    
    public int getCoreVNA() {
	return coreVNA;
    }
    public void setCoreVNA(int a) {
	coreVNA = a;
    }    
    
    public String getEdgePhysAddress () {
	return edgePhysAddress;
    }
    
    public String getEdgeProtocol () {
	return edgeProtocol;
    }
    
    public int getType () {
	return type;
    }
    public void setType(int type) {
	this.type = type;
    }
    
    public boolean evaluate (Node n) {    
	
	if (type == RESPONSE) {
	    // on the way to the dante client
	    if (n.getAddress() == coreVNA)
		return n.sendToNeighbor(this,edgeVNA);
	    else return n.deliverToApp(this,DANTE_CLIENT_PORT);
	}
	else {
	    // on the way to the dante server
	    if (n.getAddress() == coreVNA)
		return n.deliverToApp(this,DANTE_SERVER_PORT);
	    else return n.sendToNeighbor(this,coreVNA);
	}
    }
    
    public Xdr encode () {
	Xdr xdr = super.encode();
	xdr.PUT ((byte)1); // the version number;
	xdr.PUT ((short) type);
	xdr.PUT ((short) 1); // the number of requested addresses
	
	/* the refresh timer or the error code; ants ignores this value */
	xdr.PUT ((short) 30); 
	
	xdr.PUT (edgePhysAddress);
	xdr.PUT (edgeProtocol);
	xdr.PUT (edgeVNA);
	xdr.PUT (coreVNA);
	
	return xdr;
    }
    
    public Xdr decode () {
	Xdr xdr = super.decode();
	
	byte version = xdr.BYTE();
	this.type = (int) xdr.SHORT();
	
	short throwAway = xdr.SHORT();
	throwAway = xdr.SHORT();
	
	edgePhysAddress = xdr.STRING();
	edgeProtocol = xdr.STRING();
	edgeVNA = xdr.INT();
	coreVNA = xdr.INT();
	return xdr;
    }
    
    public int length () {
	return super.length() + (2 * Xdr.INT) + Xdr.STRING(edgePhysAddress) +
	    Xdr.STRING(edgeProtocol) + (3 * Xdr.SHORT) + Xdr.BYTE;
    }
    
    public DanteCapsule () {
    }
    
    public DanteCapsule (int type, int coreAddress, int edgeAddress,
			 String edgePhysAddress, String edgeProtocol) {
	
	this.edgePhysAddress = edgePhysAddress;
	this.edgeProtocol = edgeProtocol;
	this.type = type;
	this.edgeVNA = edgeAddress;
	this.coreVNA = coreAddress;
    }   
}
