/* 
 * DanteClient.java
 *
 * Copyright (c) 2000, 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.dante;

import ants.core.Application;
import ants.core.Node;
import ants.core.Capsule;
import ants.core.NeighborTable;
import ants.core.Trace;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFactory;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.AddressFormatException;

import edu.utah.janos.io.Log;


/** 
 * Sends periodic keep-alive messages to the core node in order to 
 * resubscribe to its address.
 *
 * DanteClient accepts the following arguments:
 * -node X.Y.Z.W   --- the active network address of the core node to join
 * -physical       --- the physical address of the core node
 * -protocol       --- the protocol with which to contact the core
 * -testcount	   --- number of refresh packets to send before exiting.  Testing Only.
 *
 * @author Andrew Whitaker
 */
public final class DanteClient
    extends Application 
    implements DanteConstants, Runnable {
    
    private static final String[] defaults = 
    {"-port", Integer.toString(DANTE_CLIENT_PORT)};
    
    protected Node node;
    protected NeighborTable neighbors;
    protected boolean halt;
    
    /**
     * ANTS address of the DANTE core node to attach to.
     */
    protected int coreActiveAddr = 0;
    
    /**
     * Physical address of the DANTE core node to attach to.
     */
    protected String corePhysAddr = null;
    
    /**
     * Transport protocol to use to communicate with the DANTE core
     * node.
     */
    protected String coreProtocol = null;
    
    protected final Object addressLock;
    
    /**
     * How many times we failed to connect to the server 
     */
    protected int failures = 0;
    
    /**
     * Only used for regression testing.  Causes the client to exit after
     * sending N refresh packets.
     */
    protected int testcount = -1;
    
    public DanteClient () {
	halt = false;
	addressLock = new Object();
    }
    
    public void start ()
	throws NoSuchNodeAddress, Exception {
	node = this.getNode();
	node.setAddress(0);
	neighbors = node.getNeighborTable();
	
	node.register(new DanteProtocol());
	
	Trace.trace(Trace.ANTS_DANTE,
		    "DanteClient adding tunnel to: " +Address.toString(coreActiveAddr)+
		    " (via " +coreProtocol+ " ; " +corePhysAddr+ ")");
	
	/* Add the DANTE core node as an immediate neighbor. */
	AddressFactory af = Address.getAddressFactory();
	Address link = af.newAddress(coreActiveAddr,coreProtocol,corePhysAddr);
	neighbors.addNeighbor(coreActiveAddr,link);
	
	/*
	 * Start the Dante Client thread that will keep us attached to
	 * the network.
	 */
	node.threadStart(this, "Dante Client Thread");
    }
    
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException {
	
	k.merge(defaults);
	
	for (int i = 0; i < k.length(); i++) {
	    if (k.key(i).equals("-node")) {
		try {
		    coreActiveAddr = Address.fromString(k.arg(i));
		    k.strike(i);			
		}
		catch (AddressFormatException afe) {
		    throw new InvalidKeyArgsException 
			("Address format exception " +k.arg(i)+ ": " +afe);
		}
	    }
	    
	    else if (k.key(i).equals("-physical")) {
		corePhysAddr = k.arg(i);
		k.strike(i);			
	    }		
	    else if (k.key(i).equals("-protocol")) {
		coreProtocol = k.arg(i);
		k.strike(i);			
	    }
	    else if (k.key(i).equals("-testcount")) {
		try
		{
		    testcount = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
		catch (NumberFormatException nfe)
		{
		    throw new InvalidKeyArgsException
			("Expected integer argument to -testcount:" +nfe);
		}
	    }
	}
	
	if ( (coreActiveAddr == 0)
	     || (corePhysAddr == null)
	     || (coreProtocol ==null))
	{
	    throw new InvalidKeyArgsException 
		("Must specify -node, -physical, and -protocol");
	}
	
	super.setArgs(k);
    }
    
    public void setHalt(boolean h) {
	halt = h;
    }
    
    public void receive (Capsule c) {
	if (! (c instanceof DanteCapsule)) {
	    Trace.trace(Trace.ANTS_DANTE,"Unknown capsule class");
	    return;
	}
	
	DanteCapsule cap = (DanteCapsule) c;
	if (cap.getType() != RESPONSE) {
	    Trace.trace(Trace.ANTS_DANTE,"Unknown capsule type");
	    return;
	}
	
	Trace.trace(Trace.ANTS_DANTE,"DANTE response received; " +
		    "setting address: " + Address.toString(cap.getEdgeVNA()));
	
	synchronized(addressLock) {
	    if (node.getAddress() == 0) {
		node.setAddress(cap.getEdgeVNA());
	    }
	}
    }
    
    
    public void run () {            				
	Trace.trace(Trace.ANTS_DANTE, "Staring DANTE client thread: core node: "
		    +Address.toString(coreActiveAddr)+
		    " (" +coreProtocol+ " ; " +corePhysAddr+ ")");	    
	
	Trace.trace(Trace.ANTS_DANTE, "   Client node physical address is "
		    + " (" +node.getVisibleProtocolSpec()+ " ; "
		    +node.getVisibleAddressSpec()+ ")");	    
	
	if (testcount != -1)
	{
	    Trace.trace(Trace.ANTS_DANTE,
			"Testing DANTE with testcount=" +testcount);
	}
	
	synchronized (addressLock) {
	    while (true) {
		try {
		    addressLock.wait(KEEP_ALIVE_FREQUENCY);
		}
		catch (InterruptedException _) {}
		
		if (halt)
		    return;
		
		int address = this.node.getAddress();
		
		/*
		 * This little hack is here for the DANTE regression
		 * test.
		 */
		if (testcount != -1)
		{
		    this.node.log(Log.CRIT, "DANTE testcount is " +testcount);
		    if (--testcount == 0)
			node.shutdown();
		}
		
		if (address == 0) {
		    if (++failures == MAX_FAILURES)
			this.node.shutdown();
		    
		    Trace.trace(Trace.ANTS_DANTE,"Sending DANTE 'join' message");
		    DanteCapsule dac = new DanteCapsule 
			(JOIN, coreActiveAddr, 0,
			 node.getVisibleAddressSpec(),
			 node.getVisibleProtocolSpec());	    
		    node.sendToNeighbor(dac, coreActiveAddr);
		}
		else {
		    Trace.trace(Trace.ANTS_DANTE,"Sending DANTE 'refresh' message");
		    DanteCapsule dac = new DanteCapsule 
			(REFRESH, coreActiveAddr, node.getAddress(),
			 node.getVisibleAddressSpec(),
			 node.getVisibleProtocolSpec());
		    node.send(dac);
		}
	    }
	}
    }
}
