/* 
 * DynamicRouteCapsule.java
 *  
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * Copyright (c) 2000 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.route;

import edu.utah.janos.io.Log;

import ants.core.ByteArray;
import ants.core.Capsule;
import ants.core.Node;
import ants.core.Xdr;

import java.util.Vector; // XXX
import java.util.Enumeration; // XXX

import ants.core.Assert;
import ants.core.Trace;

/**
 * Route update capsule.
 * <p>
 * Used to be a built-in capsule type.  Now it is not.
 * <p>
 * XXX fix the -accessible fields.  (Need to be accessible in DynamicRouteApp.)
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 * @author David Wetherall
 */
public class DynamicRouteCapsule
    extends Capsule
{
    final private static byte[] MID = findMID("ants.route" + ".DynamicRouteCapsule");
    protected byte[] mid()
    {
	return MID;
    }
    
    final private static byte[] PID = findPID("ants.route" + ".DynamicRouteCapsule");
    protected byte[] pid()
    {
	return PID;
    }
    
    /**
     * Physical address of the node this DynamicRouteCapsule originated at.
     * <p>
     * As a String because only the Core is allowed to parse these things.
     * XXX not very safe for pushing thorugh the network, now is it?
     */
    String protocolSpec;
    String addressSpec;
    
    /**
     * Number of route updates this capsule contains.  Controls
     * the size of the <code>routes</code> array.
     * <p>
     * XXX this need not be encoded/decoded.  It can be inferred
     * from the size of the byte array (which is already encoded
     * for us by the Xdr routines).
     */
    int nroutes;
    
    /**
     * The route updates.  Each update consists of:
     * <pre>
     * logical destination (INT)
     * cost (byte)
     * </pre>
     */
    ByteArray routes;
    
    /**
     * The length of this capsule.  Adds the length of the
     * <code>physAddr</code> string, 4 bytes for the
     * <code>nroutes</code> word, and zero or more bytes for the
     * <code>routes</code> array.
     */
    public int length()
    {
	int s = Xdr.STRING(protocolSpec) +
	    Xdr.STRING(addressSpec) +
	    Xdr.INT + // nroutes
	    ((nroutes > 0)
	     ? Xdr.BYTEARRAY(routes)
	     : 0);
	return super.length() + s;
    }
    
    public Xdr encode() 
    {
	Xdr xdr = super.encode();
	xdr.PUT(protocolSpec);
	xdr.PUT(addressSpec);
	xdr.PUT(nroutes);
	if (nroutes > 0)
	    xdr.PUT(routes);
	return xdr;
    }
    
    public Xdr decode()
    {
	Xdr xdr = super.decode();
	protocolSpec = xdr.STRING();
	addressSpec = xdr.STRING();
	nroutes = xdr.INT();
	if (nroutes > 0)
	    routes = xdr.BYTEARRAY();
	else
	    routes = null;
	return xdr;
    }
    
    /**
     * When an update capsule arrives at its destination Node, it
     * gets delivered to the app that maintains the routing table.
     */
    public boolean evaluate(Node node)
    {
	if( node.getAddress() == getDst() )
	{
	    return node.deliverToApp(this,
				     DynamicRouteApp.APP_PORT);
	}
	else
	    return node.sendToNeighbor(this, getDst());
    }  
    
    public String toString()
    {
	return "[" + super.toString()
	    + "; protocolSpec=`" + protocolSpec + "'"
	    + "; addressSpec=`" + addressSpec + "'"
	    + "; nr=" + nroutes + "]";
    }
    
    
    /**
     * Construct an empty DynamicRouteCapsule.  This is only invoked
     * when a node receives a DynamicRouteCapsule.
     */
    public DynamicRouteCapsule()
    {
    }
    
    /**
     * Create a DynamicRouteCapsule to send to a remote host.
     * This capsule will be sent to the activeDest host.
     * <p>
     * It contains the *physical address* of the host that sends
     * it.  (Because, hey, that's what routing does.)
     * <p>
     * Lastly, it will contain a suite of DynamicRouteEntries for the
     * destination host to store in its RouteTable.
     *
     * @param activeDest the destination for this capsule.  Must be 1 hop away.
     * @param protocolSpec the protocol spec for the sender's input.
     * @param addressSpec the address spec for the sender's input.
     * @param routeVec a list of updated route entries for the remote node
     */
    public DynamicRouteCapsule(int activeDest,
			       String protocolSpec,
			       String addressSpec,
			       Vector routeVec)
    {
	this.setDst(activeDest);
	this.protocolSpec = protocolSpec;
	this.addressSpec = addressSpec;
	this.nroutes = routeVec.size();
	
	if (nroutes > 0)
	{
	    routes = new ByteArray(nroutes * (Xdr.INT+Xdr.BYTE));
	    Xdr xdr = new Xdr(routes, 0);
	    for (Enumeration e = routeVec.elements(); e.hasMoreElements();)
	    {
		DynamicRouteEntry re = (DynamicRouteEntry)e.nextElement(); // XXX checkcast
		xdr.PUT(re.getLogicalDest());
		xdr.PUT((byte)re.getCost());
	    }
	}
    }
}
