/* 
 * KeyArgs.java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.util;

import ants.core.Assert;

/**
 * Keyword argument manipulation utilities.  These are used
 * for handling command line arguments.  The key is any string
 * beginning with a "-"; the argument associated
 * with the key can be any string.
 * <p>
 * <b>WARNING: Do not try to pass an argument value of "-1" to these 
 * routines.  It will be recognized as a key, not an argument.</b>
 * (Any negative number will do that, of course.)
 * <p>
 * XXX fix the above, somehow.
 *
 * @author David Wetherall
 */
public class KeyArgs
{
    /**
     * A (key, argument) pair.
     */
    private static final class KeyArg
    {
	final String key;
	String arg;
	
	/**
	 * Construct a new (key, argument) pair.
	 */
	KeyArg(String key, String arg)
	{
	    this.key = key;
	    this.arg = arg;
	}
    }
    
    /**
     * Array of KeyArgs, one for each option.
     */
    private KeyArg[] args;
    
    /**
     * The canonical 'null argument' KeyArg.
     */
    private static final KeyArg empty = new KeyArg("", null);
    
    /**
     * Return the number of key/arg pairs.
     */
    public int length()
    {
	return args.length; 
    } 
    
    /**
     * Return the i'th key.
     */
    public String key(int i)
    {
	return args[i].key;
    }
    
    /**
     * Return the i'th argument.
     */
    public String arg(int i)
    {
	return args[i].arg;
    }
    
    /**
     * Find the arg associated with the given key.  Returns
     * 'null' if the key isn't found.
     */
    public String get(String key)
    {
	for (int i = 0; i < args.length; i++)
	    if (args[i].key.equals(key))
		return args[i].arg;
	return null;
    }
    
    /**
     * Set the arg for key.  Overwrites existing
     * value for key if one is set, otherwise
     * adds a new pair to the list.
     */
    public void set(String key, String arg) {
	for (int i = 0; i < args.length; i++)
	{
	    if (args[i].key.equals(key))
	    {
		args[i].arg = arg;
		return;
	    }
	}
	
	KeyArg[] ret = new KeyArg[args.length + 1];  
	
	// XXX use System.arraycopy().
	for (int j = 0; j < args.length; j++)
	    ret[j] = args[j];
	
	ret[args.length] = new KeyArg(key, arg);
	args = ret;
    }
    
    /**
     * Strike the i'th key/arg pair from the list.
     */
    public void strike(int i)
    {
	args[i] = empty;
    }
    
    /**
     * Compact the list of key/arg pairs.  Removes all the
     * "empty" ones.
     */
    public void compact() {
	int len = 0;
	for (int i = 0; i < args.length; i++)
	    if (args[i] != empty) len++;
	
	if (len == args.length)
	    return;
	
	KeyArg[] ret = new KeyArg[len];
	int idx = 0;
	for (int i = 0; i < args.length; i++)
	    if (args[i] != empty)
		ret[idx++] = args[i];
	
	args = ret;
    }
    
    /**
     * Merge the given array of strings into the key/arg list.
     */
    public void merge(String[] ka) 
	throws InvalidKeyArgsException
    {
	merge(new KeyArgs(ka));
    }
    
    /**
     * Merge the given array of strings into the key/arg list.
     * The first <code>skip</code> entries in the string array
     * are skipped.
     */
    public void merge(String[] ka, int skip) 
	throws InvalidKeyArgsException
    {
	merge(new KeyArgs(ka, skip));
    }
    
    /**
     * Merge the given set of KeyArgs into the current list.
     * Duplicates in <code>k</code> are thrown away.
     */
    public void merge(KeyArgs k) {
	compact();
	
	if (k == null)
	    return;
	
	k.compact();
	
	// Count the duplicates
	int len = args.length;
	for (int i = 0; i < k.args.length; i++) 
	{
	    boolean match = false;
	    for (int j = 0; j < args.length; j++)
		if (args[j].key.equals(k.args[i].key)) 
		{
		    match = true;
		    break;
		}
	    if (!match)
		len++;
	}
	
	// XXX Use System.arraycopy()
	KeyArg[] ret = new KeyArg[len];  
	for (int j = 0; j < args.length; j++)
	    ret[j] = args[j];
	
	int idx = args.length;
	for (int i = 0; i < k.args.length; i++) 
	{
	    boolean match = false;
	    for (int j = 0; j < args.length; j++)
		if (args[j].key.equals(k.args[i].key))
		{
		    match = true;
		    break;
		}
	    
	    if (!match)
		ret[idx++] = k.args[i];
	}
	
	args = ret;
    }
    
    /**
     * Generate a nice, readable string representation of the key/arg
     * pairs.
     */
    public String toString()
    {
	compact();
	String ka = "";
	for (int i = 0; i < args.length; i++) {
	    ka += "(\"" + args[i].key + "\", ";
	    if (args[i].arg != null)
		ka += "\"" + args[i].arg + "\") ";
	    else
		ka += "null) ";
	}
	return ka;
    }
    
    /**
     * Generate an array of Strings, one for each
     * Key/Arg pair.
     */
    public String[] toStrings() {
	compact();
	int len = args.length;
	for (int i = 0; i < args.length; i++)
	    if (args[i].arg != null)
		len++;
	
	String[] ret = new String[len];
	
	int idx = 0;
	for (int i = 0; i < args.length; i++) {
	    ret[idx++] = args[i].key;
	    if (args[i].arg != null)
		ret[idx++] = args[i].arg;
	}
	
	return ret;
    }
    
    /**
     * A handy routine for debugging.  Throws an Error
     * if there are any key/value pairs in the current
     * KeyArgs.
     */
    public void noMore()
    {
	this.compact();
	if (this.length() > 0)
	    throw new ExtraKeyArgsError("leftover KeyArgs: " +this);
    }
    
    private KeyArgs(KeyArg[] ka) {
	args = ka;
    }
    
    public KeyArgs(String[] ka)
	throws InvalidKeyArgsException
    {
	this(ka, 0);
    }
    
    public KeyArgs() 
	throws InvalidKeyArgsException
    {
	this(null, 0);
    }
    
    public KeyArgs(String[] ka, int skip)
	throws InvalidKeyArgsException
    {
	// handle the trivial case first
	if (ka == null || ka.length == 0 || ka.length <= skip) {
	    args = new KeyArg[0];
	    return;
	}
	
	// walk the list once, counting and checking syntax
	int l = 0;
	boolean ok = false;
	for (int i = skip; i < ka.length; i++) {
	    if (ka[i] != null && ka[i].length() > 0 && ka[i].charAt(0) == '-') {
		l++;
		ok = true;
		continue;
	    }
	    if (!ok) {
		String ks = "";
		for (int j = 0; j < ka.length; j++)
		    ks += ka[j] + " ";
		throw new InvalidKeyArgsException("invalid sequence: " + ks);
	    }
	    ok = false;
	}
	
	// now fill it in for real 
	args = new KeyArg[l];
	int idx = 0;
	for (int i = skip; i < ka.length; i++) {
	    if (ka[i] != null && ka[i].length() > 0 &&  ka[i].charAt(0) == '-')
		args[idx++] = new KeyArg(ka[i], null);
	    else
		args[idx-1].arg = ka[i];
	}
    }
}
