/*
 * StatsApplication.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 */
package apps.stats;

import ants.core.Application;
import ants.core.Capsule;
import ants.core.Protocol;
import ants.core.Node;
import ants.core.NodeAddress;
import ants.core.Xdr;
import ants.core.ByteArray;
import ants.core.NeighborTable;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

import java.util.Enumeration;			// XXX
import java.util.Vector;			// XXX

/**
 * This application transverses the network, starting with the node
 * it is run from and returns statistics from all reachable nodes
 * within a specifiable depth
 * <p>
 * Parameters:
 * <ul>
 *  <li> -depth <int>: The maximum depth to go to in the tree [15].
 *  <li> -maxwait <int>: The number of seconds to wait for responses before
 *   timing out [30].
 *  <li> -senddelay <int>: The number of milliseconds to wait between sending
 *   root request capsules.  Helps decrease network strain [500].
 * </ul>
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class StatsApplication
    extends Application
    implements Runnable
{
    private final static String[] defaults = { "-depth", "15",
					       "-maxwait", "30",
					       "-senddelay", "500"};
    
    /**
     * The logging level messages issued by this class
     * are issued at.  Gross hack to hide the grosser
     * hack that is `L'.
     */
    private final int LOG_LEVEL = L[6];
    
    /**
     * The maximum depth that this request will go into
     * the network before cancelling itself (similar to TTL)
     */
    private byte depth;
    
    /**
     * Seconds to wait for activity before the run is stoped
     */
    private int maxwait;
    
    /**
     * Msec to wait between sending root request packets
     * (lessens network strain)
     */
    private int rootsenddelay;
    
    /**
     * Keeps track of the number of returned packets
     */
    private int retcount;
    
    private class SrcDestPair
    {
	public int src,dest;
	
	public SrcDestPair(int src, int dest)
	{
	    this.src=src;
	    this.dest=dest;
	}
	
	public boolean equals(Object o)
	{
	    return (((SrcDestPair)o).src==src&&
		    ((SrcDestPair)o).dest==dest);
	}
    }
    
    /**
     * Keeps track of the nodes that have sent requests that
     * have not been replied to
     */
    private Vector waitingon;
    
    /**
     * Keeps track of the parents of the nodes that have returned,
     * but that the application was not waiting on.
     * <p>
     * Sometimes return capsules come from the child before the
     * parent, this way they are not lost from the tracking system
     */
    private Vector unmatchedrets;
    
    /**
     * Handle the incoming capsule and tell it to print its information
     */
    synchronized public void receive(Capsule unknownCap)
    {
	int i,j;
	SrcDestPair sdp;
	super.receive(unknownCap);
	
	if (!(unknownCap instanceof StatsReturnCapsule))
	{
	    thisNode().log(LOG_LEVEL, "Non-stats capsule delivered!?!"+unknownCap);
	    thisNode().log(LOG_LEVEL, "StatsCap.classLoader= "+StatsCapsule.class.getClassLoader()+"; other="+unknownCap.getClass().getClassLoader());
	    return;
	}
	
	StatsReturnCapsule cap = (StatsReturnCapsule)unknownCap;
	
	cap.showInformation(thisNode());
	retcount++;
	// Keep track of what is expected to be returned
	// First add next expectations to the list
	for (i=0;i<cap.nodessentto;i++)
	{	
	    // Have I already received this one?
	    sdp=new SrcDestPair(cap.infosource,
				((Integer)cap.branches.elementAt(i)).intValue());
	    if (unmatchedrets.contains(sdp))
	    {
				// Its not unmatched now
		unmatchedrets.removeElement(sdp);
	    } else {
		waitingon.addElement(sdp);
	    }
	}
	sdp=new SrcDestPair(cap.parentnode,cap.infosource);
	// Remove what just came from the expectations
	if (waitingon.contains(sdp))
	{
	    waitingon.removeElement(sdp);
	} else {
	    unmatchedrets.addElement(sdp);
	}
	synchronized (waitingon)
	{
	    waitingon.notify();
	}
    }
    
    /**
     * Sends out the root request capsules to all
     * neighboring nodes then sleeps forever
     */
    public void run()
    {
	int lastcount;
	long lasttime;
	boolean done;
	int i;
	
	waitingon=new Vector();
	unmatchedrets=new Vector();
	thisNode().log(LOG_LEVEL, "*** Sending root stats request capsules");
	sendRootCapsules();
	
	try
	{
	    lastcount=0;
	    lasttime=thisNode().time();
	    done=false;
	    while (!done&&
		   ((waitingon.size()!=0||
		     unmatchedrets.size()!=0)||
		    retcount==0))
	    {
		lastcount=retcount;
		synchronized (waitingon)
		{
		    waitingon.wait(maxwait*1000);
		}
		if (lastcount==retcount&&thisNode().time()-lasttime>=maxwait*1000)
		    done=true;
	    }
	    if (waitingon.size()==0&&unmatchedrets.size()==0)
	    {
		thisNode().log(LOG_LEVEL,"*** Stats collection completed successfully; have a nice day.");
	    } else if (done) {
		thisNode().log(LOG_LEVEL,"*** Timeout: Stats run complete");
		for (i=0;i<waitingon.size();i++)
		{
		    thisNode().log(LOG_LEVEL,"*** Stats: "+NodeAddress.toString(((SrcDestPair)waitingon.elementAt(i)).dest)+" off of "+NodeAddress.toString(((SrcDestPair)waitingon.elementAt(i)).src)+" did not respond");
		}
	    }
	} catch (InterruptedException _) {
	}
	thisNode().shutdown();
    }
    
    /**
     * Sends the root request capsule to all neighboring nodes
     */
    private void sendRootCapsules()
    {
	NeighborTable neighbortable;
	Enumeration neighbors;
	int target;
	long keyid;
	
	neighbortable = thisNode().getNeighborTable();
	neighbors=neighbortable.getNeighbors();
	keyid=thisNode().time();
	while (neighbors.hasMoreElements())
	{
	    target=((Integer)(neighbors.nextElement())).intValue();
	    thisNode().log(LOG_LEVEL, 
			   "*** Sending root stats request (key "+
			   keyid+") to "+NodeAddress.toString(target));
	    StatsCapsule c = new StatsCapsule(target);
	    c.rootaddress=thisNode().getAddress();
	    c.keyid=keyid;
	    c.dpt=getPort();
	    c.depth=(byte)(depth+1);	// Don't count the root node
	    c.now=thisNode().time();
	    send(c);
	    // Add that destination to the expected returns
	    waitingon.addElement(new SrcDestPair(
						 thisNode().getAddress(),target));
	    try
	    {
				// Lessen network strain
		thisNode().sleep(rootsenddelay);
	    } catch (InterruptedException _) {
				// Ignore it
	    }
	}
    }
    
    /**
     * Parse command line arguments.
     * <p>
     * See the class description for a list of the arguments and
     * their meaning.
     */
    public void setArgs(KeyArgs k)
	throws InvalidKeyArgsException
    {
	int i;
	k.merge(defaults);
	
	try
	{
	    for (i=0;i<k.length();i++)
	    {
		if (k.key(i).equals("-depth"))
		{
		    depth=Byte.parseByte(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-maxwait"))
		{
		    maxwait=Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-senddelay"))
		{
		    rootsenddelay=Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
	    }
	} catch (Exception e) {
	    throw new InvalidKeyArgsException(e.getMessage());
	}
	super.setArgs(k);
    }
    
    /**
     * Register the protocol and start the thread
     */
    public void start()
	throws Exception
    {
	thisNode().register(new StatsProtocol());
	thisNode().threadStart(this);
    }
    
    /**
     * Construct a non-GUI stats app.
     * <p>
     * Doesn't do anything.
     */
    public StatsApplication()
	throws Exception
    {
    }
}

