/*
 * StatsReturnCapsule.Java
 *
 * Copyright (c) 1999, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 */
package apps.statsfusion;

import ants.core.Protocol;
import ants.core.Capsule;
import ants.core.Xdr;
import ants.core.Node;
import ants.core.ByteArray;
import ants.core.NodeAddress;
import ants.core.NodeCache;
import ants.core.NoSuchApplicationError;

import java.util.Vector;

/**
 * Stats return capsule.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class StatsReturnCapsule
    extends Capsule
{
    final private static byte[] MID = findMID("apps.statsfusion" + ".StatsReturnCapsule");
    final private static byte[] PID = findPID("apps.statsfusion" + ".StatsReturnCapsule");

    /**
     * Capsule headers and statistics returns
     */
    public int infosource;
    public int parentnode;
    public long capTime;
    public long latency;
    public long keyid;
    public int nodessentto;
    public Vector branches=new Vector();
    /**
     * Capsule headers which only occur in the main capsule
     */
    public int dpt;
    public int rootaddress;
    public ByteArray data;
    public boolean partial;
    public int numstatsstrings;
    public ByteArray statsstrings;

    /**
     * The number of seconds to wait in the NodeCache before expiring
     */
    public int cachetime;

    /**
     * An arbitrary number used to single out this capsule from other
     * return capsules in a node cache
     */
    public long removekey=0;

    /**
     * All capsules fused with this one
     */
    public Vector fusedcapsules=new Vector();

    protected byte[] mid()
    {
	return MID;
    }

    protected byte[] pid()
    {
	return PID;
    }

    public int length()
    {
	int l,i;

	l=super.length()+Xdr.INT+Xdr.INT+Xdr.INT+Xdr.BOOLEAN+Xdr.INT+Xdr.BYTEARRAY(statsstrings)+sublength();
	for (i=0;i<fusedcapsules.size();i++)
	{
	    l+=((StatsReturnCapsule)fusedcapsules.elementAt(i)).
		sublength();
	}
	return l;
    }

    /**
     * Sublength is called on each fused capsule and the parent capsule
     * to get the length of just that capsule
     */
    public int sublength()
    {
	return Xdr.INT+Xdr.INT+Xdr.LONG+Xdr.LONG+Xdr.LONG+Xdr.INT+Xdr.INT+Xdr.INT*nodessentto+Xdr.BYTEARRAY(data);
    }

    /**
     * Return capsule data looks like:<br>
     * 4	Address of the actual source of the information<br>
     * 4	Address of the node that requested stats from source node<br>
     * 8	Time capsule this is a response to was sent<br>
     * 8	Latency (0 if latency has yet to be figured)<br>
     * 8	The key id of this capsule (used in the node cache)<br>
     * 4	Number of nodes the source of this return sent to<br>
     * 4	Cachetime parameter
     * n	The addresses of the nodes the request was sent to<br>
     * n	The byte array containing the return data
     * <p>
     * The data in the capsule contains strings which specify the
     * actual statistics of the node this is returning from
     * <p>
     * The header for an entire return capsule looks like:<br>
     * 4	The number of fused capules<br>
     * 4	The destination application port<br>
     * 4	The address of the root of this tree<br>
     * 1	The partial flag (indicates that more returns might come from
     *	this source<br>
     * 4	The number of stats strings in each return capsule<br>
     * n	The return stats header strings
     */
    public Xdr encode()
    {
	int i;
	Xdr xdr = super.encode();
	xdr.PUT(fusedcapsules.size());
	xdr.PUT(dpt);
	xdr.PUT(rootaddress);
	xdr.PUT(partial);
	xdr.PUT(numstatsstrings);
	xdr.PUT(statsstrings);
	subencode(xdr);
	// Sub encode the rest
	for (i=0;i<fusedcapsules.size();i++)
	{
	    ((StatsReturnCapsule)fusedcapsules.elementAt(i)).
		subencode(xdr);
	}
	return xdr;
    }

    /**
     * Subencode is called on each fused capsule and the parent capsule
     * to encode only its own data
     */
    public void subencode(Xdr xdr)
    {
	int i;
	xdr.PUT(infosource);
	xdr.PUT(parentnode);
	xdr.PUT(capTime);
	xdr.PUT(latency);
	xdr.PUT(keyid);
	xdr.PUT(cachetime);
	xdr.PUT(nodessentto);
	for (i=0;i<nodessentto;i++)
	    xdr.PUT(((Integer)branches.elementAt(i)).intValue());
	xdr.PUT(data);
    }

    public Xdr decode()
    {
	int i,num;
	StatsReturnCapsule cap;
	Xdr xdr = super.decode();
	num=xdr.INT();
	dpt=xdr.INT();
	rootaddress=xdr.INT();
	partial=xdr.BOOLEAN();
	numstatsstrings=xdr.INT();
	statsstrings=xdr.BYTEARRAY();
	subdecode(xdr);
	for (i=0;i<num;i++)
	{
	    cap=new StatsReturnCapsule();
	    cap.subdecode(xdr);
	    // Add it
	    fusedcapsules.addElement(cap);
	}
	return xdr;
    }

    /**
     * Subencode is called on each fused capsule and the parent capsule
     * to decode only the capsules data.
     */
    public void subdecode(Xdr xdr)
    {
	int i;
	infosource=xdr.INT();
	parentnode=xdr.INT();
	capTime=xdr.LONG();
	latency=xdr.LONG();
	keyid=xdr.LONG();
	cachetime=xdr.INT();
	nodessentto=xdr.INT();
	for (i=0;i<nodessentto;i++)
	    branches.addElement(new Integer(xdr.INT()));
	data=xdr.BYTEARRAY();
    }

    public boolean evaluate(Node n)
    {
	// Check the latency
	if (latency==0)
	{
	    latency=(n.time()-capTime)/2;
	}
	if (n.getAddress()==rootaddress)
	{
	    n.log(1,"*** Delivering StatsReturnCapsule");
	    try
	    {
		return n.deliverToApp(this,dpt);
	    } catch (NoSuchApplicationError _) {
		n.log(1,"*** Caught NoSuchApplicationError; the stats root is not running");
		return false;
	    }
	} else if (n.getAddress()==infosource) {
	    // Need to route out to the destination
	    n.log(1,"*** Forwarding StatsReturnCapsule");
	    return n.routeForNode(this,getDst());
	} else {
	    if (!partial)
		n.log(1,"*** Placing StatsReturnCapsule in node cache");
	    else
		n.log(1,"*** Placing partial StatsReturnCapsule in node cache");
	    // Go sit in the node cache
	    NodeCache cache;
	    cache=n.getCache();
	    cache.put(this,this,cachetime);
	    // Wake up the stats capsule
	    synchronized (cache)
	    {
		cache.notifyAll();
	    }
	    return false;
	}
    }

    /**
     * Encode information about this node to be returned to the root
     * of the stats
     * <p>
     * To change the statistics returned by the tree, change this function
     * and the encodeHeaders function
     */
    private String[] encodeNodeInfo(Node n)
    {
	String[] retval;
	retval=new String[3];
	retval[0]=n.hostname();
	retval[1]="<unknown>";	// default to unknown
	retval[2]="<unknown>";

	try 
	{
	    retval[1]=n.getVisibleAddressSpec();
	}
	catch (Throwable securityException)
	{
	    // ignore, not allowed to see this.
	}

	try
	{
	    retval[2]=n.getVisibleProtocolSpec();
	}
	catch (Throwable securityException)
	{
	    // ignore, not allowed to see this.
	}

	return retval;
    }

/*	private ByteArray encodeNodeInfo(Node n)
	{
	ByteArray buf = new ByteArray(Xdr.STRING(n.hostname())+
	Xdr.STRING(n.getPhysicalAddressSpec())+
	Xdr.STRING(n.getPhysicalProtocolSpec()));
	Xdr xdr = new Xdr(buf,0);
	xdr.PUT(n.hostname());
	xdr.PUT(n.getPhysicalAddressSpec());
	xdr.PUT(n.getPhysicalProtocolSpec());
	return buf;
	}*/

    /**
     * Encode the stats header strings.  These must be one-to-one with the
     * strings returned by the encodeNodeInfo function
     */
    private String[] encodeHeaders()
    {
	String[] retval;
	retval=new String[3];
	retval[0]="Hostname";
	retval[1]="Physical address spec";
	retval[2]="Physical protocol spec";
	return retval;
    }

/*	private ByteArray encodeHeaders()
	{
	ByteArray buf = new ByteArray(Xdr.INT+
	Xdr.STRING("Hostname")
	Xdr.STRING("Physical address spec")+
	Xdr.STRING("Physical protocol spec"));
	Xdr xdr = new Xdr(buf,0);
	xdr.PUT(3);		// Indicates three items of info
	xdr.PUT("Hostname");
	xdr.PUT("Physical address spec");
	xdr.PUT("Physical protocol spec");
	return buf;
	}*/

    /**
     * Convert the return of encodeHeaders() to a byte array which can
     * be encoded
     */
    private ByteArray nodeHeadersToByteArray()
    {
	String nodeheaders[];
	ByteArray buf;
	Xdr xdr;
	int i,l;
	nodeheaders=encodeHeaders();
	numstatsstrings=nodeheaders.length;
	l=0;
	for (i=0;i<nodeheaders.length;i++)
	    l+=Xdr.STRING(nodeheaders[i]);
	buf=new ByteArray(l);
	xdr=new Xdr(buf,0);
	for (i=0;i<nodeheaders.length;i++)
	    xdr.PUT(nodeheaders[i]);
	return buf;
    }

    /**
     * Print out the information that is contained in this capsule
     */
    public void showInformation(StatsReturnCapsule hsrc,Node n)
    {
	Xdr xdr=new Xdr(data,0);
	Xdr xdrheaders;
	int i;
	if (hsrc==null)
	    hsrc=this;
	xdrheaders=new Xdr(hsrc.statsstrings,0);
	n.log(0, "*** Recieved stats return capsule from "+
	      NodeAddress.toString(infosource)+
	      " with parent "+NodeAddress.toString(parentnode)+
	      " with latency of "+latency+" containing:");
	for (i=0;i<hsrc.numstatsstrings;i++)
	    n.log(0, "\t"+xdrheaders.STRING()+": "+xdr.STRING());
	// Print fused capsules
	for (i=0;i<fusedcapsules.size();i++)
	    ((StatsReturnCapsule)fusedcapsules.elementAt(i)).
		showInformation(this,n);
    }

    /**
     * Fuses the information from src into this capsule
     * <p>
     * Note that this only holds on to a reference to the capsule
     * to be fused until this capsule is actually sent
     * <p>
     * @param src - The capsule to fuse with this one
     */
    public void fuse(StatsReturnCapsule src)
    {
	int i;
	fusedcapsules.addElement(src);
	// Need to fuse in capules that one
	for (i=0;i<src.fusedcapsules.size();i++)
	{
	    fusedcapsules.addElement(src.fusedcapsules.elementAt(i));
	}
	src.fusedcapsules.removeAllElements();
    }

    public void setData(ByteArray ba)
    {
	data=ba;
    }

    /**
     * Used when a StatsReturnCapsule is held in a NodeCache
     */
    public boolean equals(Object o)
    {
	// The infosource is also used so that a return capsule can
	// be uniquely identified in a node cache, but also allows
	// for finding unknown returns in a node cache if the object
	// being compared to has a infosource of 0
	if ((o instanceof StatsReturnCapsule)&&
	    ((StatsReturnCapsule)o).keyid==keyid&&
	    ((StatsReturnCapsule)o).removekey==removekey&&
	    (((StatsReturnCapsule)o).infosource==infosource||
	     ((StatsReturnCapsule)o).infosource==0))
	    return true;
	return false;
    }

    /**
     * Used when a StatsReturnCapsule is held in a NodeCache
     */
    public int hashCode()
    {
	return (int)keyid;
    }

    public StatsReturnCapsule()
    {
    }

    public StatsReturnCapsule(int d)
    {
	setDst(d);
    }

    /**
     * Constructor used when StatsReturnCapsule is sitting in the
     * node cache
     * <p>
     * This sets only the information neccessary to compare this capsule
     * to those sitting in the node cache.
     */
    public StatsReturnCapsule(long keyid, int rootaddress)
    {
	this.keyid=keyid;
	this.rootaddress=rootaddress;
    }

    public StatsReturnCapsule(short sa, short da, int na, Node n)
    {
	String nodeinfo[];
	ByteArray buf;
	Xdr xdr;
	int i,l;
	dpt=da;
	setDst(na);
	nodeinfo=encodeNodeInfo(n);
	// Build the xdr for the node info
	l=0;
	for (i=0;i<nodeinfo.length;i++)
	    l+=Xdr.STRING(nodeinfo[i]);
	buf=new ByteArray(l);
	xdr=new Xdr(buf,0);
	for (i=0;i<nodeinfo.length;i++)
	    xdr.PUT(nodeinfo[i]);
	setData(buf);
	statsstrings=nodeHeadersToByteArray();
    }
}
